"""This is version 2 of the framework, a simplified but more extensive version
It's purpose is to leave alot of the work up to the user in the GUI but also create
a basic simulated version of your application. It will:
- Gather a list of GAM datasources
- Iterate through all states and remove signals produced by external datasources
  replacing these with constantGAMs
- Iterate through all states and find all produced signals
- Create an IOGAM and FileWriter Datasource for these and add this to every state, thread 0.
- For each state, it will iterate through this and create a constantGAM for any producer
  signals not produced in said state.
- Among these constantGAMs it will add a State integer which will increment for each state.
- When the function returns it will provide a dictionary of state name to state integer value.

Req:
- It will support multi-layered types
- It should support between threads using the RealTimeThreadAsyncBridge:
    - In order to do this we need to for each state, identify what signals are not present 
      in thread 0 but are present in other threads.
    - It will then create a AsyncBridge between the two threads to bring that 
      threads' signals into thread 0.
"""
import copy

from martepy.functions.extra_functions import getname # pylint: disable=W0614
from martepy.functions.gam_functions import (getAlias,
                                             isMatchingSignal,
                                             findSignalInFunctionList,
                                             getKeyAttribute,
                                             setDatasource,
                                             addDimensions)
from martepy.marte2.datasources.timing_datasource import TimingDataSource
from martepy.marte2.objects.genericobject import MARTe2Object
from martepy.marte2.gams import IOGAM, ConstantGAM
from martepy.marte2.datasources.linux_timer import LinuxTimer
from martepy.marte2.datasources.gam_datasource import GAMDataSource
from martepy.marte2.datasources.files.reader import FileReader
from martepy.frameworks.end_gam import EndGAM
from martepy.marte2.datasources.async_bridge import AsyncBridge

class SimulationGenerator():
    ''' This class is the simulation generator, it takes a pre-defined MARTe2Application instance,
    replaces the input signals from the outside world and logs all signals which occur during
    execution.
    
    *** It is a highly complex set of functions as it achieves a complex task - as a result, we've
    pylint disabled a few check regarding complex code but we're not exceeding those parameters
    too much and most of these have thorough commenting in. ***'''
    def __init__(self, app, configure=True):
        """Initialise our simulation generator object
        """
        self.original_app = app
        self.reset()
        self.constants_gam_source = GAMDataSource('+' + self.constants_ddb.lstrip('+'))
        self.state_constant_vals = {}
        self.gam_sources = []
        self.replaced_datasources = []
        if configure:
            # Configure immediately
            self.configure()

    def reset(self):
        """Reset our application
        """
        # Reset CPU Configurator
        self.simulation_app = copy.deepcopy(self.original_app)
        self.simulation_app.types_used = []
        self.log_ddb = 'LogGAMSource'
        self.constants_ddb = 'ConstantGAMDatasource'
        self.misc_ddb = 'OutsideWorldGAMDDB'
        self.gam_sources = []
        self.overridden_signal_names = []
        self.simulation_app.config['maxcycles'] = 500

    def configure(self, app=None):
        """This configures the simulation by iterating the application
        and looking for all datasources, for each one it then qualifies
        if they are replacable. If they are then they are catalogued.
        """
        # Allow reseting the app here
        if app is not None:
            self.original_app = app
            self.reset()

        # Collect our GAM Datasources
        self.gam_sources = [a.configuration_name.lstrip('+') for a in
                            self.original_app.additional_datasources if isinstance(
                                a, GAMDataSource)]

        # We want to keep in LinuxTimer's for the user
        linux_timers = [a for a in self.original_app.additional_datasources if isinstance(
            a, LinuxTimer)]

        # # Strip out non-GAM Datasources except: AsyncBridge, FileReader.
        # datasources = self.simulation_app.additional_datasources
        # datasources = [a for a in self.original_app.additional_datasources if isinstance(a, (
        #     AsyncBridge, FileReader, GAMDataSource, TimingDataSource))] + linux_timers

        self._internal_datasources = [a.configuration_name.lstrip('+') for a in
                                 self.original_app.additional_datasources if isinstance(a, (
                                     AsyncBridge, FileReader, GAMDataSource, TimingDataSource))]

        self._internal_datasources += [self.misc_ddb, self.log_ddb, self.constants_ddb]
        self._internal_datasources += [a.configuration_name.replace('+','') for a in linux_timers]
        # Ensure we know what we removed
        self.replaced_datasources = [a for a in self.original_app.additional_datasources if not (
            isinstance(a, (AsyncBridge, FileReader, GAMDataSource, LinuxTimer, TimingDataSource)))]

    def _createGamSources(self):
        ''' Define the GAMSources that are used to partition signals '''
        # Create the GAM DataSource for holding our constants and signals
        self.constants_gam_source = GAMDataSource('+' + self.constants_ddb.lstrip('+'))
        LogGAMSource = GAMDataSource('+' + self.log_ddb.lstrip('+'))
        miscDDB = GAMDataSource('+' + self.misc_ddb.lstrip('+'))
        # Ensure we're not reading this
        for item in [self.constants_gam_source, miscDDB, LogGAMSource]:
            if item not in self.simulation_app.additional_datasources:
                self.simulation_app.additional_datasources += [item]

    def _handleTimer(self):
        ''' Function to either inject a timer and return it or find the timer
        block if already defined by the user and return a pointer to it '''
         # We want to give the user the ability to add and use their own timers whilst simulating
        timer = next((a for a in self.simulation_app.additional_datasources if isinstance(
            a, LinuxTimer)), None)
        freq = self.simulation_app.config['timefrequency']
        if not timer:
            timer = LinuxTimer(
                        configuration_name = '+Timer',
                        cpu_mask = 12,
                        sleep_nature = 'Busy',
                        execution_mode = 'RealTimeThread',
                        frequency = freq
                    )
            self.simulation_app.add(additional_datasources = [timer])

        return timer, freq

    def _handlerTimerIO(self, timer, freq):
        # Find the function reading from our linux timer
        timing_block = None

        for state in self.simulation_app.states:
            if state.threads.objects:
                thread = state.threads.objects[0]
                timer_name = timer.configuration_name.replace('+','')
                timing_block = next((a for a in thread.functions if any(
                    b for b in a.input_signals if b[1]['MARTeConfig']['DataSource'] == timer_name)
                    ),None)
                # Only one IOGAM can read from the linuxtimer,
                if not timing_block:
                    timing_block = IOGAM(
                            configuration_name = '+TimerHandler',
                            input_signals = [
                                ('Time', {'MARTeConfig': {'DataSource': 'Timer',
                                                           'Type': 'uint32', 
                                                           'Frequency': freq}}),
                            ],
                            output_signals = [
                                ('DTime', {'MARTeConfig': {'DataSource': self.gam_sources[0],
                                                            'Alias': 'DTime', 'Type': 'uint32'}}),
                            ],)
                    if timing_block not in self.simulation_app.functions:
                        self.simulation_app.functions.insert(0, timing_block)
                    if timing_block not in thread.functions:
                        thread.functions.insert(0, timing_block)
                else:
                    # Find which signal specifies the Frequency
                    freq_input = next((a for a in timing_block.input_signals if 'Frequency' in
                                       list(a[1]['MARTeConfig'].keys())), None)
                    if freq_input:
                        freq_input[1]['MARTeConfig']['Frequency'] = freq
                    else:
                        timing_block.input_signals[0][1]['MARTeConfig']['Frequency'] = freq

        return timing_block

    def _createEndStateSignals(self):
        ''' This function section creates the state signal and end-signal:
        Only works if a thread was created for this state '''

        for state in self.simulation_app.states:
            maxcycles = self.simulation_app.config['maxcycles']
            if len(state.threads.objects) > 0:
                state_name = state.configuration_name.lstrip('+')
                thread = state.threads.objects[0]
                thread_name = thread.configuration_name.lstrip('+')
                state_signal = [('state', {'MARTeConfig': {'DataSource': self.gam_sources[0],
                                                           'Type': 'uint32'}})]
                endgam = EndGAM(f'+{state_name}-{thread_name}-end',int(maxcycles),state_signal)
                thread.functions += [endgam]
                self.simulation_app.functions += [endgam]

        # Create a constant into thread0 of each state that allows us to maintain some
        # tracking of which state we are in when we look at logs later on
        self.state_constant_vals = {}
        for index, state in enumerate(self.simulation_app.states):
            state_name = getname(state)
            Statecon = ConstantGAM(f'+{state_name}ConstantLog',
                                   [('state', {'MARTeConfig': {'Type': 'uint32',
                                                               'Default': index,
                                                               'DataSource':
                                                               self.gam_sources[0]}})])
            if len(state.threads.objects) > 0:
                self.simulation_app.functions += [Statecon]
                state.threads.objects[0].functions.insert(0,Statecon)
            self.state_constant_vals[state_name] = index

    def _replaceExternalSignals(self, timing_block): # pylint: disable=R0914, R0912
        ''' This function works out what signals come from the outside world and replaces them '''
        # Only IOGAM's can go between Datasources
        ninternal_datasources = [getname(a) for a in self.replaced_datasources]
        self.simulation_app.additional_datasources = [a for a in
                                                      self.simulation_app.additional_datasources if
                                                      getname(a) not in ninternal_datasources]
        for s_i, state in enumerate(self.simulation_app.states): # pylint: disable=R1702
            state_name = getname(state)
            for t_i, thread in enumerate(state.threads.objects):
                # Iterate through each state and each subsequent thread
                thread_name = getname(thread)
                # Catalogue the functions we add, maintain a boolean if we added any
                # and create the first constant GAM
                new_functions = []
                top_constant_gam = ConstantGAM(f'+{state_name}{thread_name}TopLvlConstants',[])
                added=False
                for f_i, function in enumerate(thread.functions):
                    # For each function, if it is an IOGAM
                    curr_func = self.simulation_app.states[s_i].threads.objects[t_i].functions[f_i]
                    if isinstance(function, IOGAM):
                        # Ignore our timing block as we don't want to replace this.
                        if function == timing_block:
                            continue # Don't mess with the timer_handler
                        # Is a transferring function possibly
                        to_pop = []
                        for input_signal in function.input_signals:
                            # Check each input signal, if they come from non-managed datasources
                            # they come from the outside world and need replacing
                            datasource = input_signal[1]['MARTeConfig']['DataSource']
                            if datasource in ninternal_datasources:
                                # We need to replace this signal
                                # Manage if it's an EPICS signal
                                if 'PVName' in input_signal[1]['MARTeConfig'].keys():
                                    del input_signal[1]['MARTeConfig']['PVName']

                                # We can't do inplace for new functions to the
                                # thread so we append afterwards. Now delete the input and
                                # it's corresponding output signal.
                                to_pop += [input_signal] # can't remove them
                                # within an iterator of them.
                        # Now that we have found what input signals we need to replace
                        # Remove the IO Input signal we replaced and the output signal
                        for itemtoremove in to_pop:
                            # Figure out the function
                            function = curr_func
                            signals_list = function.input_signals
                            # Where our signal is
                            index = signals_list.index(itemtoremove)
                            added=True
                            # Get the output signal - this is what we need to really replace with a
                            # default value
                            output_signal = function.output_signals[index]
                            alias = getAlias(output_signal)
                            # Track what signals we override
                            self.overridden_signal_names.insert(0,alias)
                            out_type = output_signal[1]['MARTeConfig']['Type']
                            default = 0 if out_type in ('uint32','uint64', 'uint8') else 0.0
                            # Now replace the signal and insert it into our constantsGAM
                            io_gam = self._replaceSignal(output_signal, top_constant_gam,
                                                getname(self.constants_gam_source),
                                                self.simulation_app, default)
                            # Append our IOGAM if one was needed to be created
                            if io_gam:
                                new_functions += [io_gam]
                            # Now remove the signal now that it is done
                            signals_list.remove(itemtoremove)
                            function.output_signals.pop(index)
                        for o_i, output_signal in enumerate(function.output_signals):
                            datasource = output_signal[1]['MARTeConfig']['DataSource']
                            if datasource in ninternal_datasources:
                                # This feeds to the outside world - we don't want to do this,
                                # we should place it into our misc DDB
                                qsignal = function.output_signals[o_i]
                                qsignal[1]['MARTeConfig']['DataSource'] = self.misc_ddb

                if added:
                    # If we needed to replace signals, now we can add our constant GAM as it
                    # now contains signals
                    thread.functions.insert(0,top_constant_gam)
                    for new_function in new_functions:
                        # Any other new functions also need adding to the thread definition
                        thread.functions.insert(1,new_function)

    def build(self): # pylint: disable=R0914, R0912
        """This function is where the actual work gets done to create the new application
        for simulation, configure simply collects information based on the given application
        instance.
        """
        self._createGamSources()

        timer, freq = self._handleTimer()

        timing_block = self._handlerTimerIO(timer, freq)

        self.simulation_app.default_data_source = self.gam_sources[0]

        self._createEndStateSignals()

        self._replaceExternalSignals(timing_block)

        # Resanitize an application between its state, threads functions
        # and it's functions definitions.
        self.simulation_app.sanitize()

        self.simulation_app.buildLog()

        # Now we iterate through again and use a multitude of any's to check:
        # Is there in this state, any thread, with an IOGAM, that is producing that
        # required signal into our LogDDB
        for state in self.simulation_app.states: # pylint: disable=R1702
            missingsignals = ConstantGAM(f'+missingsignals{getname(state)}', [])
            missingio = IOGAM(f'+missingiosignals{getname(state)}',[],[])
            # This gets flagged up by pylint as unnecessary but is in fact very necessary!
            # It delinks the pointer refs
            total_signals = [a for a in self.simulation_app.logging_iogam.input_signals] # pylint: disable=R1721
            total_signals += [a for a in self.simulation_app.async_to_io.input_signals] # pylint: disable=R1721
            for signal in total_signals:
                # This is a complicated if any setup, we're asking, is any signal of any function
                # in any thread of our current state, matching our signal we are looking for
                if len(state.threads.objects) > 0:
                    # Is a used thread
                    if not any(
                        any(any(isMatchingSignal(output_signal, signal) for
                                output_signal in function.output_signals) for
                                function in thread.functions) for thread in state.threads.objects):
                        # There is no producer for this signal within this state
                        # use replace signal to create this signal in thread 0 for this state
                        # We need to account for the fact that defaults must match if this was
                        # produced by a constantGAM for the same datasource.
                        # So we must therefore find the original producer of this signal
                        producer = findSignalInFunctionList(signal, self.simulation_app.functions)

                        default = 0
                        if producer: # a Producer was found
                            if 'Default' in producer[1]['MARTeConfig'].keys(): # It has a default
                                default = getKeyAttribute(producer, 'Default')

                        modsignal = copy.deepcopy(signal)
                        # This next step might seem redundant but it was found
                        # necessary during testing - we ensure we don't repeat the name in
                        # our missing signals gam
                        alias = getAlias(modsignal)
                        unique_name = self.simulation_app.newUniqueName(alias)
                        if not any(out_signal[0] == unique_name for
                                   out_signal in missingsignals.output_signals):
                            # Don't duplicate our work
                            modsignal = (unique_name,modsignal[1])
                            modsignal[1]['MARTeConfig']['Alias'] = unique_name
                            setDatasource(modsignal, self.constants_ddb)
                            self._replaceSignal(modsignal, missingsignals,
                                                self.constants_ddb, self.simulation_app, default)
                            missingio.input_signals += [modsignal]
                            missingio.output_signals += [signal]

            if len(missingsignals.output_signals) > 0:
                state.threads.objects[0].functions += [missingsignals, missingio]

        self.simulation_app.sanitize()

        self.simulation_app.removeUnused()

        # Now we want to insert the internals of types we don't have in the application
        # Need to insert these into the objects part of the application so they appear
        # at the very start
        if self.simulation_app.types_used:
            for type_name in self.simulation_app.types_used:
                empty_obj = MARTe2Object(f'+{type_name}', f'{type_name}::{type_name}Cls')
                self.simulation_app.add(externals=[empty_obj])
        return self.simulation_app

    def unflattenReplace(self, signal_to_replace, type_name, # pylint: disable=R0914
                          constant_gam=None, constantgamsource=None):
        ''' This is a recursive algorithm to unflatten a signal and create a Constant
        GAM which replicates the signal with it's defaults and then combines it back to it's
        complex type structure. '''
        # Get the type definition
        ualias = getAlias(signal_to_replace)
        type_def = self.simulation_app.type_db.getTypeByName(type_name)
        if type_def.name not in self.simulation_app.types_used:
            self.simulation_app.types_used += [type_def.name]
        # For each struct member
        field_members = []
        for field in type_def.fields:
            # Get a name for it
            alias = field.name
            # If it's not fundamental, unpack it as well
            if not self.simulation_app.type_db.isFundamental(field.type):
                # Recursively unpack it
                unflattened_signal = self.unflattenReplace(self, field.type, constantgamsource)
                field_members += unflattened_signal.output_signals[0]
            else:
                # Create the constant value for this field member
                # It's output name should match the same convention of how unflatten
                # creates the overall output name
                constant_signal = (type_name+alias,{'MARTeConfig':{}})
                constant_signal[1]['MARTeConfig']['Type'] = field.type

                # Set the default value for our replacement signal
                default = 0 if field.type in ('uint32', 'uint64', 'uint8') else 0.0
                # We need to account for if our type is multiple elements
                if 'NumberOfElements' in constant_signal[1]['MARTeConfig']:
                    num_elements = int(getKeyAttribute(constant_signal,'NumberOfElements'))
                    # Okay then create the default array format
                    default = "{" + ' '.join([str(default)]*num_elements) + "}"
                    addDimensions(constant_signal) # Double check we have dimensions too
                else:
                    addDimensions(constant_signal)
                constant_signal[1]['MARTeConfig']['Default'] = str(default)
                constant_signal[1]['MARTeConfig']['DataSource'] = constantgamsource
                constant_signal[1]['MARTeConfig']['Alias'] = type_name+alias+ualias
                constant_signal = (type_name+alias,constant_signal[1])
                constant_gam.output_signals += [constant_signal]

                # Now our field member exists, we need to construct our type
                field_member_for_io = copy.deepcopy(constant_signal)
                if 'Default' in field_member_for_io[1]['MARTeConfig']:
                    del field_member_for_io[1]['MARTeConfig']['Default']
                field_members += [field_member_for_io]

        output_signal_for_io = copy.deepcopy(signal_to_replace)
        type_io_gam = IOGAM(f'+IO{ualias}_{type_name}', field_members,[output_signal_for_io])
        self.simulation_app.functions += [type_io_gam]
        return type_io_gam

    def _replaceSignal(self, signal_to_replace=(), constant_gam=None, constantgamsource=None, # pylint: disable=R0917
                       application=None, default=0):
        """This function will create the constant that replaces this signal.
        If this is a basic type, it will simply add this to the constant_gam
        and then add that to the iogam to get filtered into our constantGAMSource,
        however if it is a more complex signal type, it will create the necessary
        constants and IOGAM's to construct this signal type and then place this in
        the IOGAM, the new functions will be added to thread.functions and
        application.functions in place by reference.
        """
        constant_signal = copy.deepcopy(signal_to_replace)
        if 'Frequency' in list(constant_signal[1]['MARTeConfig'].keys()):
            del constant_signal[1]['MARTeConfig']['Frequency']
        type_name = constant_signal[1]['MARTeConfig']['Type']
        # Get the type and check if fundamental
        if not application.type_db.isFundamental(type_name):
            # Really we need to follow the type
            # Challenge: How to handle Types which we may not know in advance?
            # Maybe we make this recursively support nested types? or just keep it flat and simple
            # Start by creating a ConstantGAM from extracting the type
            top_lvl = self.unflattenReplace(signal_to_replace,
                                             type_name, constant_gam, constantgamsource)

            return top_lvl
        # Is a fundamental type
        # Very simple, just add the type and default 0 to our constantGAM
        alias = getAlias(signal_to_replace)
        num_elements = 1
        if 'NumberOfElements' in list(constant_signal[1]['MARTeConfig'].keys()):
            num_elements = int(getKeyAttribute(constant_signal,'NumberOfElements'))
            default = "{" + ' '.join([str(default)]*num_elements) + "}"
        addDimensions(constant_signal)
        new_signal = (signal_to_replace[0], {'MARTeConfig':{
            'Type': type_name, 'Default': str(default),
            'DataSource': constant_signal[1]['MARTeConfig']['DataSource'],
            'Alias': alias}})
        new_signal[1]['MARTeConfig']['NumberOfElements'] = str(num_elements)
        if 'NumberOfDimensions' in list(constant_signal[1]['MARTeConfig'].keys()):
            num_dim = int(getKeyAttribute(constant_signal,'NumberOfDimensions'))
            new_signal[1]['MARTeConfig']['NumberOfDimensions'] = str(num_dim)
        else:
            new_signal[1]['MARTeConfig']['NumberOfDimensions'] = '1'

        constant_gam.output_signals += [new_signal]
        return None



''' Easy method to import all the HTTP Objects '''

from martepy.marte2.objects.http.directoryresource import MARTe2HttpDirectoryResource
from martepy.marte2.objects.http.messageinterface import MARTe2HttpMessageInterface
from martepy.marte2.objects.http.objectbrowser import MARTe2HTTPObjectBrowser
from martepy.marte2.objects.http.service import MARTe2HttpService

__all__ = [
    'MARTe2HttpDirectoryResource',
    'MARTe2HttpMessageInterface',
    'MARTe2HTTPObjectBrowser',
    'MARTe2HttpService',
]



"""This is the reader capable of reading a MARTe2 cfg file and translating this into
the MARTe2 Pythonic class representations
"""
import copy
import os

from martepy.marte2.generic_application import GAMDataSource, MARTe2Application
from martepy.marte2.objects import (MARTe2Message,
                                    MARTe2ReferenceContainer,
                                    MARTe2StateMachine,
                                    MARTe2StateMachineEvent)
from martepy.marte2.objects.http.directoryresource import MARTe2HttpDirectoryResource
from martepy.marte2.objects.http.messageinterface import MARTe2HttpMessageInterface
from martepy.marte2.objects.http.service import MARTe2HttpService
from martepy.marte2.objects.real_time_state import MARTe2RealTimeState
from martepy.marte2.objects.real_time_thread import MARTe2RealTimeThread
from martepy.marte2.objects.gam_scheduler import MARTe2GAMScheduler
from martepy.marte2.factory import Factory as mpyFactory
from martepy.marte2.objects.http.objectbrowser import MARTe2HTTPObjectBrowser
from martepy.marte2.objects.configuration_database import MARTe2ConfigurationDatabase
from martepy.marte2.qt_functions import genNextStateMsgs
from martepy.functions.extra_functions import getname

class TreeNode:
    ''' Node which represents an item in the configuration for navigating the tree '''
    def __init__(self, name, parent = None):
        self.name = name
        self.parameters = {}
        self.children = []
        self.parent = parent

    def addParameter(self, key, value):
        ''' Add a parameter to the config item '''
        self.parameters[key] = value

    def addChild(self, child):
        ''' Add a child to the config tree item '''
        self.children.append(child)

    def __str__(self, level=0):
        ''' Display the item details as a string as it would in the cfg '''
        result = "\t" * level + f"{self.name}\n"
        for key, value in self.parameters.items():
            result += "\t" * (level + 1) + f"{key} = {value}\n"
        for child in self.children:
            result += child.__str__(level + 1)
        return result


def parseFile(file_path):
    ''' Parse a cfg file given a path - wrapper around the parse text function '''
    with open(file_path, 'r', encoding='utf-8') as file:
        content = file.read()
    return content


def buildTree(content):
    ''' Read a text content and parse it into TreeNode classes for later interpretation '''
    root = TreeNode('root')
    current_node = root
    lines = content.split('\n')
    building = False
    current_line = ''
    key = ''
    for line in lines:
        stripped_line = line.lstrip(' ')
        if building:
            current_line += stripped_line
            if not stripped_line.strip().endswith('\\n'):
                building = False
                current_node.addParameter(key, current_line.replace('\\n','\n'))
            continue
        if stripped_line.strip().endswith('{'):
            # Start of a new TreeNode
            name = stripped_line.split('=')[0].strip()
            new_node = TreeNode(name, current_node)
            current_node.addChild(new_node)
            current_node = new_node
        elif stripped_line.strip().endswith("}"):
            if '=' in stripped_line and '+' not in stripped_line:
                key, value = stripped_line.split("=")
                key = key.strip()
                value = value.strip()
                current_node.addParameter(key, value)
                continue
            current_node = current_node.parent
        elif "=" in stripped_line:
            key, value = stripped_line.split("=")
            key = key.strip()
            value = value.strip()
            if stripped_line.strip(' ').endswith('\\n'):
                building = True
                current_line = value
            else:
                current_node.addParameter(key, value)

    return root


def formatToSignal(tree_definition):
    '''Need to change to the signal format:
    signals = [('AnalogueIn',{'MARTeConfig':{
            'Alias': 'n/c',
            'DataSource': self.datasource,
            'Type': 'float64',
            'NumberOfElements': 1
        }})]
    '''
    if tree_definition:
        signals = []
        for child in tree_definition.children:
            name = child.name
            parameters = {}
            for param in list(child.parameters):
                key = param
                value = child.parameters[key]
                parameters[key] = value.replace('"','')
            signals.append((name, {'MARTeConfig':parameters}))
        return signals
    return []

def getSignals(function_def, formatter):
    ''' Given a cfg definition of a signal - return this into the signal format
    we universally use '''
    return formatToSignal(next((a for a in function_def.children if a.name == formatter), None))

def readApplication(file_path):
    ''' Read an application given a file path - wrapper around the parse file function '''
    file_content = parseFile(file_path)
    return readApplicationText(file_content)

class UnrecognisedParameterException(Exception):
    ''' Exception that should be thrown when an unknown parameter in a tree occurs '''
    def __init__(self, class_name, key):
        self.msg = f'Unrecognised parameter caught for class {class_name} and parameter {key}'

def getParameters(child):
    ''' Return the parameters of an object '''
    return {key:value for key, value in child.parameters.items() if key != 'Class'}

def setParameters(child, child_obj, class_name):
    ''' Set an objects parameters '''
    for key, value in getParameters(child).items():
        if hasattr(child_obj, key.lower()):
            setattr(child_obj, key.lower(), value.replace('"','').replace("'",''))
        else:
            raise UnrecognisedParameterException(class_name,key)

def handleChildObjects(parent_obj, function, factory):
    ''' Handle when a function contains objects of it's own - so far only needed 
    for the MessageGAM that we know of but could be applicable to other GAMs. '''
    if function.children:
        # So far we've only seen this with MessageGAM but it gives the concept that
        # there may be objects within a function
        for child in function.children:
            if child.name in ('InputSignals', 'OutputSignals'):
                continue
            # Something unknown, we should check that our gam has a read_gam_config function
            # It should have a class attribute, so we can use that to decipher what that class is
            # It's name should also correspond to its attribute name in the GAM in lower case so
            # we know where to put it. This should be true recursively down.
            # But, if the class does not have attr, we will dump this into the objects attribute -
            # assuming it goes there.
            if child.name[0] == '+':
                # Is a new object class to build
                class_name = child.parameters['Class']
                child_cls = factory.create(class_name)
                child_obj = child_cls(configuration_name=child.name)
                if hasattr(child_obj, 'objects'):
                    child_obj.objects = []
                if class_name == 'ConfigurationDatabase':
                    # We make an exception for the configuration database - may regret later.
                    child_obj.objects = getParameters(child)
                else:
                    setParameters(child, child_obj, class_name)
                handleChildObjects(child_obj, child, factory)
                if hasattr(parent_obj, child.name.strip('+').lower()):
                    setattr(parent_obj, child.name.strip('+').lower(), child_obj)
                else:
                    parent_obj.objects += [child_obj]
            elif hasattr(parent_obj, child.name.strip('+').lower()):
                setattr(parent_obj, child.name.strip('+').lower(), child.parameters)
            else:
                parent_obj.objects += [child_obj]

def getRootClass(tree_root, class_name):
    ''' Iterator that returns the first node in a given tree with a matching class '''
    for result in tree_root.children:
        if hasattr(result, 'parameters'):
            # Written this way to make more readable - initially was a one line list comprehension
            if any(param == 'Class' for param in list(result.parameters)):
                if result.parameters['Class'] == class_name:
                    return result
    # Nothing was found, return None like you would with next(({condition}), None)
    return None

mfactory = mpyFactory()
main_dir = os.path.abspath(os.path.dirname(__file__))
mfactory.loadRemote(os.path.join(main_dir,"objects","objects.json"))
mfactory.loadRemote(os.path.join(main_dir,"gams","gams.json"))
mfactory.loadRemote(os.path.join(main_dir,"datasources","datasources.json"))
mfactory.loadRemote(os.path.join(os.path.dirname(main_dir),"frameworks","end.json"))

def toFunctionObj(function):
    ''' Convert object to it's function class '''
    class_name = function.parameters['Class']
    block_cls = mfactory.create(class_name)
    input_signals = getSignals(function, "InputSignals")
    output_signals = getSignals(function, "OutputSignals")
    configuration_name = function.name.strip('+')
    blk = block_cls(configuration_name=configuration_name, input_signals=input_signals,
                    output_signals=output_signals)
    for parameter_name, parameter_value in function.parameters.items():
        if parameter_name == 'Class':
            continue
        setattr(blk, parameter_name.lower(), parameter_value.replace('"',''))
    handleChildObjects(blk, function, mfactory)
    return configuration_name, blk

def toDataSourceObj(datasource):
    ''' Convert object to it's datasource class '''
    configuration_name = datasource.name.lstrip('+')
    class_name = datasource.parameters['Class']
    signals = getSignals(datasource, "Signals")
    try:
        blk_cls = mfactory.create(class_name)
    except ValueError:
        blk_cls = mfactory.create(configuration_name)
    try:
        blk = blk_cls(configuration_name=configuration_name, output_signals=signals)
    except TypeError:
        blk = blk_cls(configuration_name=configuration_name, input_signals=signals)
    # Catch any leftover parameters
    param_cpy = copy.deepcopy(datasource.parameters)
    del param_cpy['Class']
    for param in list(param_cpy):
        value = param_cpy[param].strip('"').strip("'")
        setattr(blk, param.lower(), value)
    return blk

def returnFromChildren(children, name):
    ''' Given a tree nodes children, return the named node or None '''
    for child in children:
        if child.name == name:
            return child
    return None

def readApplicationText(file_content):
    ''' Given text content, read this in and interpret it into the MARTe2
    Pythonic class representations '''
    app = MARTe2Application()
    tree_root = buildTree(file_content)
    # The first set of objects should be anything root level and our application.
    # First let's find our application, we can worry about
    # root level objects later when we incorporate those
    application_definition = getRootClass(tree_root, 'RealTimeApplication')
    app.app_name = application_definition.name.lstrip('$')

    # Okay now we have this, there will be four child sections:
    # - Functions
    # - Data (DataSources)
    # - States
    # - Scheduler
    function_map = {}
    # Now iterate our functions and interpret them based on our factory knowledge of them
    for function in returnFromChildren(application_definition.children, '+Functions').children:
        # There will always be a class parameter here
        configuration_name, blk = toFunctionObj(function)
        app.add(functions=[blk])
        function_map[configuration_name.strip('+')] = blk

    # Handle the datasource section
    datasources = returnFromChildren(application_definition.children, '+Data')

    for datasource in datasources.children:
        blk = toDataSourceObj(datasource)
        app.add(additional_datasources=[blk])

    # Handle if we have a defined default DataSource
    if 'DefaultDataSource' in list(datasources.parameters.keys()):
        app.default_data_source = datasources.parameters['DefaultDataSource']
    else:
        if len([a for a in app.additional_datasources if isinstance(a, GAMDataSource)]) == 0:
            app.default_data_source = 'DDB0'
        else:
            app.default_data_source = [a for a in app.additional_datasources if
                                       isinstance(a, GAMDataSource)][0]

    # Get the states defined in our application
    getStates(application_definition, function_map, app)
    # And now the scheduler finally

    scheduler = next(a for a in application_definition.children if a.name == '+Scheduler')
    app.add(internals=[
        MARTe2GAMScheduler(configuration_name = scheduler.name,
                           timing_datasource_name=scheduler.parameters['TimingDataSource'],
                           class_name=scheduler.parameters['Class'])])

    new_state_machine = defineStateMachine(tree_root, application_definition,
                                           app, getname(app.states[0]))

    app.add(externals=[new_state_machine])

    found_http_browser, found_http_service, http_messages = toHttp(tree_root)
    if found_http_browser:
        app.add(externals=[found_http_browser])
    if found_http_service:
        app.add(externals=[found_http_service])
    return app, new_state_machine, found_http_browser, http_messages

def getStates(application_definition, function_map, app):
    ''' Get an applications states definition '''
    # Now decipher the states
    states = next(a for a in application_definition.children if a.name == '+States')
    for state in states.children:
        config_name = state.name
        threads = MARTe2ReferenceContainer('Threads', objects=[])
        for thread in state.children[0].children:
            thread_name = thread.name
            cpus = thread.parameters['CPUs'] if 'CPUs' in list(thread.parameters.keys()) else 0x1
            if 'x' in cpus:
                cpus = int(float.fromhex(cpus))
            else:
                cpus = int(cpus)
                functions = '{ }'
            if 'Functions' in list(thread.parameters.keys()):
                functions = thread.parameters['Functions']
            functions = functions.strip('{').strip('}').split(' ')
            functions = [function_map[function] for function in functions if function != '']
            # Find the actual function we evaluated with this name
            new_thread = MARTe2RealTimeThread(configuration_name = thread_name,
                                              cpu_mask = cpus, functions = functions)
            threads.objects.append(new_thread)
        new_state = MARTe2RealTimeState(configuration_name = config_name, threads = threads)
        app.add(states = [new_state])

    # Ensure we have an error state with a thread definition
    error_state = next((a for a in app.states if a.configuration_name == '+Error'), None)
    if error_state:
        if not error_state.threads.objects:
            new_thread = MARTe2RealTimeThread(configuration_name = '+Thread1',
                                              cpu_mask = 4294967295, functions = [])
            error_state.threads.objects = [new_thread]
    else:
        threads = MARTe2ReferenceContainer('Threads', objects=[])
        threads.objects = [MARTe2RealTimeThread(configuration_name = '+Thread1',
                                          cpu_mask = 4294967295, functions = [])]
        new_state = MARTe2RealTimeState(configuration_name = '+Error', threads = threads)
        app.add(states = [new_state])

def defineStateMachine(tree_root, application_definition, app, first_state): # pylint:disable=R0914
    ''' Find our state machine, if not found, create one'''
    state_machine = getRootClass(tree_root, 'StateMachine')
    if state_machine is None:
        return createBasicStateMachine(application_definition, app, first_state)
    new_state_machine = MARTe2StateMachine(state_machine.name, [])
    for states in state_machine.children:
        new_state = MARTe2ReferenceContainer(states.name, [])
        for event in states.children:
            if event.parameters['Class'] == 'StateMachineEvent':
                timeout = -1
                try:
                    # Won't necessarily be defined for Error Events
                    timeout = event.parameters['Timeout'].strip('"')
                except KeyError:
                    pass
                new_event = MARTe2StateMachineEvent(event.name, event.parameters['NextState'],
                                                    event.parameters['NextStateError'],
                                                    timeout, [])
            else:
                new_event = MARTe2ReferenceContainer('+ENTER', [])
            for message in event.children:
                maxwait = -1
                try:
                    maxwait = message.parameters['MaxWait'].strip('"')
                except KeyError:
                    pass
                new_message = MARTe2Message(message.name,
                                            message.parameters['Destination'].strip('"'),
                                            message.parameters['Function'].strip('"'),
                                            MARTe2ConfigurationDatabase(objects={}),
                                            message.parameters['Mode'],
                                            maxwait)
                if message.children:
                    parameter = message.children[0]
                    objects = {key:value for key, value in
                                parameter.parameters.items() if key != 'Class'}
                    new_message.parameters = MARTe2ConfigurationDatabase(parameter.name,
                                                                            objects=objects)
                if event.parameters['Class'] == 'StateMachineEvent':
                    new_event.messages += [new_message]
                else:
                    new_event.objects += [new_message]
            new_state.objects += [new_event]
        new_state_machine.states += [new_state]
    return new_state_machine

def createBasicStateMachine(application_definition, app, first_state):
    ''' creates a very basic state machine object '''
    # Create a state machine basic version based on the number of states
    states = []
    # First you want an entry point which just starts the application
    app_name = application_definition.name
    msgs = genNextStateMsgs(first_state.replace('+',''), app_name)
    msgs.pop(0)
    initial_event = MARTe2StateMachineEvent('+START',
                                            first_state.upper().replace('+',''),
                                            'ERROR', 0, msgs)
    states += [MARTe2ReferenceContainer("+INITIAL", [initial_event])]
    msgs = genNextStateMsgs('Error', app_name)
    state1 = MARTe2StateMachineEvent('+ERROR', 'ERROR', 'ERROR', 0, [])
    states += [MARTe2ReferenceContainer("+STATE1", [state1])]

    # The error state is special, it needs the enter and reset state,
    # the enter state must exist!
    msgs = genNextStateMsgs('Error', app_name)
    error_enter = MARTe2ReferenceContainer('+ENTER', msgs)
    msgs = genNextStateMsgs(first_state, app_name)
    error_reset = MARTe2StateMachineEvent('+RESET', first_state.upper().replace('+',''),
                                            first_state.upper().replace('+',''), 0, msgs)
    states += [MARTe2ReferenceContainer("+ERROR", [error_enter, error_reset])]

    # Check that an error state exists, if not, create one for the application
    if not any(getname(state).upper() == "ERROR" for state in app.states):
        threads = MARTe2ReferenceContainer('Threads', objects=[])
        threads.objects = [MARTe2RealTimeThread(configuration_name = '+Thread1',
                                          cpu_mask = 4294967295, functions = [])]
        new_state = MARTe2RealTimeState(configuration_name = '+Error', threads = threads)
        app.add(states=[new_state])

    new_state_machine = MARTe2StateMachine('+StateMachine', states)
    return new_state_machine

def createHttpBrowser(http_browser, http_messages):
    ''' From our given found browser, create it's object '''
    if http_browser is None:
        return None
    parent_browser = MARTe2HTTPObjectBrowser(http_browser.name,
                                                http_browser.parameters['Root'],
                                                objects=[])
    object_browser = getRootClass(http_browser, 'HttpObjectBrowser')
    if object_browser:
        attributes = [object_browser.name, object_browser.parameters['Root']]
        parent_browser.objects += [MARTe2HTTPObjectBrowser(*attributes,
                                                            objects=[])]
        # This is the initial child version,
        # we're going to build the children objects first
    resources_html = getRootClass(http_browser, 'HttpDirectoryResource')
    if resources_html:
        attributes = [resources_html.name, resources_html.parameters['BaseDir']]
        parent_browser.objects += [MARTe2HttpDirectoryResource(*attributes)]
    message_interface = getRootClass(http_browser, 'HttpMessageInterface')
    if message_interface:
        if message_interface.children:
            # Has messages defined
            for message in message_interface.children:
                maxwait = int(message.parameters['MaxWait'].strip('"'))
                new_message = MARTe2Message(message.name,
                                            message.parameters['Destination'].strip('"'),
                                            message.parameters['Function'].strip('"'),
                                            MARTe2ConfigurationDatabase(objects={}),
                                            message.parameters['Mode'].strip('"'),
                                            maxwait)
                if message.children:
                    # Has a parameter set defined
                    parameter = message.children[0]
                    objects = {key:value for key, value in
                                parameter.parameters.items() if key != 'Class'}
                    new_message.parameters = MARTe2ConfigurationDatabase(parameter.name,
                                                                            objects=objects)
                http_messages.append(new_message)
        parent_browser.objects += [MARTe2HttpMessageInterface(resources_html.name,
                                                                http_messages)]
    return parent_browser

def toHttp(tree_root):
    ''' Find in our application the HTTP objects'''
    found_http_browser = None
    found_http_service = None
    http_messages = []
    try:
        http_browser = getRootClass(tree_root, 'HttpObjectBrowser')
        http_service = getRootClass(tree_root, 'HttpService')
        # First interpret the WebRoot
        if http_browser:
            found_http_browser = createHttpBrowser(http_browser, http_messages)
        # Now interpret the WebService
        if http_service:
            attributes = [http_service.name,
                          http_service.parameters['Port'],
                          http_service.parameters['WebRoot'],
                          http_service.parameters['Timeout'],
                          http_service.parameters['ListenMaxConnections'],
                          http_service.parameters['AcceptTimeout'],
                          http_service.parameters['MaxNumberOfThreads'],
                          http_service.parameters['MinNumberOfThreads']]
            found_http_service = MARTe2HttpService(*attributes)
    except AttributeError:
        http_browser = None # Will handle later on in the call stack upwards
        http_service = None
    return found_http_browser, found_http_service, http_messages

if __name__ == "__main__":
    CFG_PATH = "Simulation.cfg"
    main_dir = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
    test_file = os.path.join(main_dir, 'tests','frameworks',
                             'simulation_framework','multistatethread_complextype.cfg')
    appcfg = readApplication(test_file)[0]
    with open('Out.cfg','w', encoding='utf-8') as outfile:
        outfile.write(appcfg.writeToConfig())



''' Pythonic representation of the Expression GAM'''

from functools import partial

from PyQt5.QtWidgets import (
    QTextEdit,
    QLabel
)

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import addInputSignalsSection, addOutputSignalsSection, paraChange
class ExpressionGAM(MARTe2GAM):
    ''' Pythonic representation of the Expression GAM'''
    def __init__(self,
                    configuration_name: str = 'ExpressionGAM',
                    input_signals: list = [],
                    output_signals: list = [],
                    expression: str = '',
                ):
        self.expression = expression
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'MathExpressionGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )

    def writeGamConfig(self, config_writer):
        ''' Write the GAM configuration - i.e. the expression '''
        expression = self.expression.split('\n')
        firstline = expression[0]
        tabs = config_writer.tab_text * config_writer.tab
        tabs += ' ' * len('Expression = ')
        expression = [tabs + a for a in expression]
        expression[0] = firstline
        expression = '\n'.join(expression)
        config_writer.writeNode('Expression', f'"{expression}"')

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['parameters']['expression'] = self.expression
        res['parameters']['Class name'] = 'ExpressionGAM'
        res['label'] = "ExpressionGAM"
        res['block_type'] = 'ExpressionGAM'
        res['class_name'] = 'ExpressionGAM'
        res['title'] = f"{self.configuration_name} (ExpressionGAM)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to our class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.expression = data['parameters']["expression"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)


        # Define Expression text
        lbl_wgt = QLabel("Expression: ")
        expression_wgt = QTextEdit(mainpanel_instance)
        expression_wgt.setPlainText(node.parameters['expression'])
        expression_wgt.textChanged.connect(partial(paraChange, expression_wgt, node, 'expression'))
        mainpanel_instance.configbarBox.addWidget(lbl_wgt, 2, 0)
        mainpanel_instance.configbarBox.addWidget(expression_wgt, 2, 1)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("ExpressionGAM", ExpressionGAM, plugin_datastore)
    factory.registerBlock("MathExpressionGAM", ExpressionGAM, plugin_datastore)



''' Pythonic representation of ReferenceContainer '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2Object(MARTe2ConfigObject):
    ''' Pythonic representation of ReferenceContainer '''
    def __init__(self,
                    configuration_name: str = 'Message',
                    class_name = "ReferenceContainer",
                    objects = [],
                ):
        super().__init__()
        self.class_name = class_name
        self.objects = objects
        self.configuration_name = configuration_name

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        config_writer.startClass('+' + getname(self), self.class_name)
        for obj in self.objects:
            obj.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res["objects"] = [a.serialize() for a in self.objects]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data['class_name']
        def toObj(a):
            return factory.create(a['class_name'])().deserialize(a, factory=factory)
        # This part is tricky as a ref container can
        # contain any object so the factory needs to recircle here
        self.objects = [toObj(a) for a in data["objects"]]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("ReferenceContainer", MARTe2Object, plugin_datastore)
    factory.registerBlock("MARTe2Object", MARTe2Object, plugin_datastore)



''' Pythonic representation of the PID GAM'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import (addInputSignalsSection,
                                         addOutputSignalsSection,
                                         addLineEdit)

class PIDGAM(MARTe2GAM):
    ''' Pythonic representation of the PID GAM'''
    def __init__(self,
                    configuration_name: str = 'PID',
                    input_signals: list = [],
                    output_signals: list = [],
                    kp = 0.0,
                    ki = 0.0,
                    kd = 0.0,
                    samplefrequency = 1.0,
                    maxoutput = 1.0,
                    minoutput = 0.0
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'PIDGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )
        self.kp = kp
        self.ki = ki
        self.kd = kd
        self.samplefrequency = samplefrequency
        self.maxoutput = maxoutput
        self.minoutput = minoutput

    def writeGamConfig(self, config_writer):
        ''' Write our GAM Configuration '''
        config_writer.writeNode("kp",self.kp)
        config_writer.writeNode("ki",self.ki)
        config_writer.writeNode("kd",self.kd)
        config_writer.writeNode("sampleFrequency",self.samplefrequency)
        config_writer.writeNode("maxOutput",self.maxoutput)
        config_writer.writeNode("minOutput",self.minoutput)

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['parameters']['kp'] = self.kp
        res['parameters']['ki'] = self.ki
        res['parameters']['kd'] = self.kd
        res['parameters']['samplefrequency'] = self.samplefrequency
        res['parameters']['maxoutput'] = self.maxoutput
        res['parameters']['minoutput'] = self.minoutput
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to the class instance '''
        super().deserialize(data, hashmap, restore_id)
        # Now we build up
        self.kp = data['parameters']["kp"]
        self.ki = data['parameters']["ki"]
        self.kd = data['parameters']["kd"]
        self.samplefrequency = data['parameters']["samplefrequency"]
        self.maxoutput = data['parameters']["maxoutput"]
        self.minoutput = data['parameters']["minoutput"]
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)

        addLineEdit(mainpanel_instance, node, "Kp: ", 'kp', 3, 0)

        addLineEdit(mainpanel_instance, node, "Ki: ", 'ki', 3, 2)

        addLineEdit(mainpanel_instance, node, "Kd: ", 'kd', 4, 0)

        addLineEdit(mainpanel_instance, node, "Sample Frequency: ", 'samplefrequency', 4, 2)

        addLineEdit(mainpanel_instance, node, "Max Output: ", 'maxoutput', 5, 0)

        addLineEdit(mainpanel_instance, node, "Min Output: ", 'minoutput', 5, 2)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("PIDGAM", PIDGAM, plugin_datastore)



''' Pythonic representative class of the EPICS Subscriber Datasource '''
from martepy.marte2.datasource import MARTe2DataSource

from martepy.marte2.qt_functions import addLineEdit, addOutputSignalsSection

class EPICSSubscriber(MARTe2DataSource):
    ''' Pythonic representation of the EPICS Subscriber class '''
    def __init__(self,
                    configuration_name: str = 'EPICSSubscriber',
                    input_signals = [],
                    output_signals = [],
                    StackSize = 1048576,
                    CPUs = 0xff,
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'EPICSCA::EPICSCAInput',
                output_signals = output_signals,
            )
        self.stacksize = StackSize
        self.cpus = CPUs

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('StackSize', f'{self.stacksize}')
        if isinstance(self.cpus, str):
            if 'x' in self.cpus:
                config_writer.writeNode('CPUs', self.cpus)
            else:
                config_writer.writeNode('CPUs', hex(int(self.cpus)))
        else:
            config_writer.writeNode('CPUs', hex(self.cpus))
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the EPICS Subscriber'''
        res: dict = super().serialize()
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['cpus'] = self.cpus
        res['label'] = "EPICSSubscriber"
        res['block_type'] = 'EPICSSubscriber'
        res['class_name'] = 'EPICSSubscriber'
        res['title'] = f"{self.configuration_name} (EPICSSubscriber)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the EPICS Subscriber '''
        res = super().deserialize(data, hashmap, restore_id)
        self.stacksize = data['parameters']["stacksize"]
        self.cpus = data['parameters']["cpus"]
        return res

    def writeSignals(self, defs, config_writer):
        """Use MARTe1ConfigObject.writeSignals but remove any DataSources."""
        omit_keys = ('DataSource','Alias')
        signals = [(
                n,
                dict(d, **{'MARTeConfig': {
                        k: v for k, v in d['MARTeConfig'].items() if k not in omit_keys
                    }})
            ) for n, d in defs]

        super().writeSignals(signals, config_writer)

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addOutputSignalsSection(mainpanel_instance, node, epics=True)

        if isinstance(node.parameters['cpus'], str):
            if 'x' in node.parameters['cpus']:
                node.parameters['cpus'] = node.parameters['cpus']
            else:
                node.parameters['cpus'] = hex(int(node.parameters['cpus']))
        else:
            node.parameters['cpus'] = hex(node.parameters['cpus'])

        addLineEdit(mainpanel_instance, node,"CPU mask: ",'cpus', 2, 0)

        addLineEdit(mainpanel_instance, node,"StackSize: ",'stacksize', 2, 2)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("EPICSSubscriber", EPICSSubscriber, plugin_datastore)
    factory.registerBlock("EPICSCA::EPICSCAInput", EPICSSubscriber, plugin_datastore)



''' The basic definition for a MARTe2 python datasource object '''

import logging
import copy
from martepy.marte2.config_object import MARTe2ConfigObject

logger = logging.getLogger(__name__)

class MARTe2DataSource(MARTe2ConfigObject):
    """Object for configuring GAMs for MARTe2RealTimeThread instances"""

    plugin = 'marte2_datasources'
    def __init__(self,
                    configuration_name: str = 'datasource',
                    class_name: str = 'UnknownDataSource',
                    input_signals: list = [],
                    output_signals: list = [],
                ):
        super().__init__(configuration_name=configuration_name)
        self.class_name = class_name
        self.comment = ''
        self.input_signals = list(input_signals)
        self.output_signals = list(output_signals)

    def writeDatasourceConfig(self, config_writer):
        ''' Needs implementing but how to write the datasource '''
        raise NotImplementedError()

    def writeInputSignals(self, config_writer, section_name='InputSignals'):
        ''' Write the input signals section '''
        if self.input_signals:
            config_writer.startSection(section_name)
            self.writeSignals(self.input_signals, config_writer)
            config_writer.endSection(section_name)

    def writeOutputSignals(self, config_writer, section_name='OutputSignals'):
        ''' Write the output signal section '''
        if self.output_signals:
            config_writer.startSection(section_name)
            self.writeSignals(self.output_signals, config_writer)
            config_writer.endSection(section_name)

    def write(self, config_writer):
        ''' Default method of writing the datasource - assuming you have written the
        writeDatasourceConfig '''
        config_writer.startClass('+' + self.configuration_name.strip('+'), self.class_name)
        self.writeDatasourceConfig(config_writer)
        config_writer.endSection('+' + self.configuration_name.strip('+'))

    def writeSignals(self, defs, config_writer): # pylint: disable=W0237, W0221
        """Use MARTe1ConfigObject.writeSignals but remove any DataSources."""
        super().writeSignals([(
                n,
                dict(d, **{'MARTeConfig': {
                        k: v for k, v in d['MARTeConfig'].items() if
                        k not in ('DataSource','Alias')
                    }})
            ) for n, d in defs], config_writer)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            myself = copy.deepcopy(self.serialize())
            theother = copy.deepcopy(other.serialize())
            myself['id'] = 0
            theother['id'] = 0
            if myself == theother:
                return True
        return False

    def serialize(self):
        ''' Serialize the datasource '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['configuration_name'] = self.configuration_name
        res['inputsb'] = self.input_signals
        res['outputsb'] = self.output_signals
        res['class_name'] = self.class_name
        res['block_type'] = self.class_name
        res['comment'] = self.comment
        res['rank'] = False
        res['plugin'] = 'marte2_datasources'
        res['label'] = self.class_name
        res['parameters'] = {}
        res['content'] = {}
        res['inputs'] = []
        res['outputs'] = []
        res['id'] = id(self)
        res['pos_x'] = 0
        res['title'] = self.configuration_name + "(" + self.class_name + ")"
        # Need to have this match the format expected in GUI
        res['pos_y'] = 0
        return copy.deepcopy(res)

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool: # pylint: disable=W0221
        ''' Deserialize our object from a dictionary '''
        super().deserialize(data, hashmap, restore_id)
        #self.class_name = data['class_name']
        self.configuration_name = "+" + data['configuration_name'].strip('+')
        self.input_signals = data["inputsb"]
        self.output_signals = data["outputsb"]
        self.configuration_name = data['configuration_name']
        self.comment = data['comment']
        return self



''' Pythonic representative class of the LinuxTimer Datasource '''

import copy

from PyQt5.QtWidgets import (
    QWidget,
    QSizePolicy
)

from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addLineEdit

class LinuxTimer(MARTe2DataSource):
    ''' Pythonic representation of the LinuxTimer '''
    def __init__(self,
                    configuration_name: str = 'Timer',
                    sleep_nature: str = 'Default',
                    execution_mode: str = 'IndependentThread',
                    cpu_mask: int = 0xFFFFFFFF,
                    sleep_percentage: int = 0,
                    phase: int = 0,
                    frequency = 1000,
                    input_signals = [],
                    output_signals = []
                ):
        self.frequency = frequency
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'LinuxTimer'
            )
        self.sleep_nature = sleep_nature
        self.execution_mode = execution_mode
        self.cpu_mask = cpu_mask
        self.frequency = frequency
        self.phase = phase
        self.sleep_percentage = sleep_percentage

    def _defineBasicSignal(self, name, marte_type):
        ''' Simplification of defining the output signals '''
        return (name,{'MARTeConfig':{'Type': marte_type,
                                     'NumberOfDimensions': '1', 'NumberOfElements':'1'}})

    @property
    def output_signals(self):
        ''' Return all the possible signals always '''
        frequency_sgl = self._defineBasicSignal('Time', 'uint32')
        frequency_sgl[1]['MARTeConfig']['Frequency'] = self.frequency
        return [self._defineBasicSignal('Counter', 'uint32'),
                frequency_sgl,
                self._defineBasicSignal('AbsoluteTime', 'uint64'),
                self._defineBasicSignal('DeltaTime', 'uint64'),
                self._defineBasicSignal('TrigRephase', 'uint8')]

    @output_signals.setter
    def output_signals(self, _):
        ''' Enforce that we always have all output signals available and defined '''

    def writeDatasourceConfig(self, config_writer):
        ''' Write the configuration of our linux timer '''
        config_writer.writeNode('SleepNature', f'"{self.sleep_nature}"')
        config_writer.writeNode('ExecutionMode', f'"{self.execution_mode}"')
        if isinstance(self.cpu_mask, str):
            if 'x' in self.cpu_mask:
                config_writer.writeNode('CPUMask', self.cpu_mask)
            else:
                config_writer.writeNode('CPUMask', hex(int(self.cpu_mask)))
        else:
            config_writer.writeNode('CPUMask', hex(self.cpu_mask))
        config_writer.writeNode('Phase', f'{self.phase}')
        config_writer.writeNode('SleepPercentage', f'{self.sleep_percentage}')
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize our timer object '''
        res = super().serialize()
        res['parameters']['sleep_nature'] = self.sleep_nature
        res['parameters']['execution_mode'] = self.execution_mode
        res['parameters']['cpu_mask'] = self.cpu_mask
        res['parameters']['frequency'] = self.frequency
        res['parameters']['sleep_percentage'] = self.sleep_percentage
        res['parameters']['phase'] = self.phase
        output_signals = []
        for output in self.output_signals:
            newoutput = copy.deepcopy(output)
            newoutput[1]['MARTeConfig']['Alias'] = newoutput[0]
            output_signals.append(newoutput)
        res['outputsb'] = output_signals
        res['outputs'] = [(0, 'Counter'),
                          (0, 'Time'),
                          (0, 'AbsoluteTime'),
                          (0, 'DeltaTime'),
                          (0, 'TrigRephase')]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserilize the timer parameters '''
        super().deserialize(data, hashmap, restore_id)
        self.sleep_nature = data['parameters']["sleep_nature"]
        self.execution_mode = data['parameters']["execution_mode"]
        self.cpu_mask = data['parameters']["cpu_mask"]
        self.frequency = data['parameters']['frequency']
        # introduce backwards compatibility with versions prior to allowing user
        # configuration of sleep percentage and phase
        try:
            self.sleep_percentage = data['parameters']['sleep_percentage']
        except KeyError:
            self.sleep_percentage = 0
        try:
            self.phase = data['parameters']['phase']
        except KeyError:
            self.phase = 0
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        ''' Loads the configuration panel used in the XMARTe2 GUI '''
        # Fix to make it look better
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        mainpanel_instance.configbarBox.addWidget(spacer, 0, 2)
        addLineEdit(mainpanel_instance, node, 'Frequency: ','frequency', 1, 0)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        mainpanel_instance.configbarBox.addWidget(spacer, 1, 2)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our object with the factory '''
    factory.registerBlock("LinuxTimer", LinuxTimer, plugin_datastore)



''' Pythonic representation of RealTimeState '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.marte2.objects.referencecontainer import MARTe2ReferenceContainer
from martepy.functions.extra_functions import getname

def genericThread():
    ''' Return an empty Ref Container '''
    return MARTe2ReferenceContainer(configuration_name='Threads',objects=[])

class MARTe2RealTimeState(MARTe2ConfigObject):
    """Object for configuring RealTimeStates for MARTe2 applications"""

    def __init__(self,
                    configuration_name: str = '+State',
                    threads: MARTe2ReferenceContainer = genericThread(),
                ):
        super().__init__()
        self.configuration_name = configuration_name
        self.class_name = 'RealTimeState'
        self.threads = threads

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        config_writer.startClass('+' + getname(self), self.class_name)
        self.threads.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['threads'] = self.threads.serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.threads = factory.create('ReferenceContainer')().deserialize(data["threads"],
                                                                          factory=factory)
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("RealTimeState", MARTe2RealTimeState, plugin_datastore)



''' Easy import of all available GAMs '''

from martepy.marte2.gams.iogam import IOGAM
from martepy.marte2.gams.constant_gam import ConstantGAM
from martepy.marte2.gams.ssm_gam import SSMGAM
from martepy.marte2.gams.conversion import ConversionGAM
from martepy.marte2.gams.expression_gam import ExpressionGAM
from martepy.marte2.gams.message_gam import MessageGAM
from martepy.marte2.gams.mux import MuxGAM
from martepy.marte2.gams.filter import FilterGAM
from martepy.marte2.gams.pid import PIDGAM
from martepy.marte2.gams.waveform_chirp import WaveformChirpGAM
from martepy.marte2.gams.waveform_points import WaveformPointsGAM
from martepy.marte2.gams.waveform_sin import WaveformSinGAM

__all__ = [
        'IOGAM',
        'ConstantGAM',
        'SSMGAM',
        'ConversionGAM',
        'ExpressionGAM',
        'MessageGAM',
        'MuxGAM',
        'FilterGAM',
        'PIDGAM',
        'WaveformChirpGAM',
        'WaveformPointsGAM',
        'WaveformSinGAM'
    ]



''' Pythonic representative class of the AsyncBridge Datasource '''

import copy
from functools import partial

from PyQt5.QtWidgets import (
    QLabel,
    QComboBox
)

from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addSignalsSection, addLineEdit


class AsyncBridge(MARTe2DataSource):
    ''' Pythonic representation of the AsyncBridge '''
    def __init__(self,
                    configuration_name: str = 'RealTimeThreadAsyncBridge',
                    numbuffers = 1,
                    heapname = 'Default',
                    blocking_mode = 0,
                    resetmsec_timeout = -1,
                    input_signals = [],
                    output_signals = [],
                    inputs = False
                ):
        super().__init__(
                 configuration_name = configuration_name,
                 class_name = 'RealTimeThreadAsyncBridge',
                 input_signals = input_signals,
            )
        if not inputs:
            self.output_signals = self.input_signals
            self.input_signals = []
        self.numbuffers = numbuffers
        self.heapname = heapname
        self.blocking_mode = blocking_mode
        self.resetmsec_timeout = resetmsec_timeout
        self.input = inputs

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration to the cfg. '''
        # Based on experience and padova examples, prescibing any information to the datasource
        # results in exceptions
        # config_writer.writeNode('NumberOfBuffers', '"{}"'.format(self.numbuffers))
        # config_writer.writeNode('HeapName', '"{}"'.format(self.heapname))
        # config_writer.writeNode('BlockingMode', '"{}"'.format(self.blocking_mode))
        # config_writer.writeNode('ResetMSecTimeout', '"{}"'.format(self.resetmsec_timeout))

    def serialize(self):
        ''' Serialize the datasource object '''
        res: dict = super().serialize()
        res['parameters']['numbuffers'] = self.numbuffers
        res['parameters']['heapname'] = self.heapname
        res['parameters']['blocking_mode'] = self.blocking_mode
        res['parameters']['resetmsec_timeout'] = self.resetmsec_timeout
        res['parameters']['input'] = self.input

        # Override name
        res['label'] = "AsyncBridge"
        res['block_type'] = 'AsyncBridge'
        res['class_name'] = 'AsyncBridge'
        res['title'] = f"{self.configuration_name} (AsyncBridge)"
        return res

    # We need to override this to make an exception for the fact that we have
    # self.input for GUI purposes but actually this is the same datasource
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            myself = copy.deepcopy(self.serialize())
            theother = copy.deepcopy(other.serialize())
            if not myself['configuration_name'] == theother['configuration_name']:
                return False
            del myself['parameters']['input']
            del theother['parameters']['input']
            if not myself['parameters'] == theother['parameters']:
                return False
            return True
        return False

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the datasource object. '''
        res = super().deserialize(data, hashmap, restore_id)
        self.numbuffers = data['parameters']["numbuffers"]
        self.heapname = data['parameters']["heapname"]
        self.input = data['parameters']['input']
        self.blocking_mode = data['parameters']["blocking_mode"]
        self.resetmsec_timeout = data['parameters']["resetmsec_timeout"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node): # pylint: disable=R0915
        """This function is intended to be for the GUI where it can call the static
        instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        # This needs to be specialised, if the user selects to have this reading or writing
        # Then the sockets need to be switched

        def removeWidgetsExceptRow0(layout):
            for i in reversed(range(layout.count())):
                item = layout.itemAt(i)
                row_span = layout.getItemPosition(i)[3]
                column_span = layout.getItemPosition(i)[2]
                if row_span == 1 and column_span == 1:
                    layout.removeItem(item)
                    widget = item.widget()
                    if widget is not None:
                        widget.setParent(None)
                    del item

        def inputChanged(wgt_field, mainpanel_instance, node, _):
            # Switch socket positions
            prev = node.scene.large_import
            node.scene.large_import = True
            merged_list = node.inputsb + node.outputsb
            type_input = False
            if wgt_field.currentText() == 'Read':
                node.inputsb = merged_list
                node.outputsb = []
                type_input = True
                for socket in node.outputs:
                    for edge in socket.edges:
                        edge.remove()
                sockets = copy.copy(node.outputs)
                for socket in sockets:
                    socket.is_input = True
                    socket.is_output = False
                    socket.position = 2
                    socket.node = node
                    socket.grSocket.setParentItem(None)
                    socket.node.scene.grScene.removeItem(socket.grSocket)
                    del socket.grSocket
                    socket.grSocket = socket.__class__.Socket_GR_Class(socket)
                    socket.setSocketPosition()
                node.inputs = sockets
                node.outputs = []
            else:
                # Is write
                node.inputsb = []
                node.outputsb = merged_list
                for socket in node.inputs:
                    for edge in socket.edges:
                        edge.remove()
                sockets = copy.copy(node.inputs)
                for socket in sockets:
                    socket.is_input = False
                    socket.is_output = True
                    socket.position = 5
                    socket.node = node
                    socket.grSocket.setParentItem(None)
                    socket.node.scene.grScene.removeItem(socket.grSocket)
                    del socket.grSocket
                    socket.grSocket = socket.__class__.Socket_GR_Class(socket)
                    socket.setSocketPosition()
                node.outputs = sockets
                node.inputs = []

            node.parameters['input'] = type_input
            node.updateSocketPositions()
            node.scene.large_import = prev
            # Redraw
            draw(mainpanel_instance, node, type_input)

        type_input = node.parameters['input']

        def draw(mainpanel_instance, node, type_input):
            removeWidgetsExceptRow0(mainpanel_instance.configbarBox)
            addSignalsSection(mainpanel_instance, node, type_input=type_input)

            wgt_label = QLabel("Read/Write: ")
            wgt_field = QComboBox(mainpanel_instance)
            wgt_field.addItems(['Read', 'Write'])
            index = 0 if type_input else 1
            wgt_field.setCurrentIndex(index)
            wgt_field.currentIndexChanged.connect(partial(inputChanged, wgt_field,
                                                          mainpanel_instance, node))
            mainpanel_instance.configbarBox.addWidget(wgt_label, 2, 0)
            mainpanel_instance.configbarBox.addWidget(wgt_field, 2, 1)

            addLineEdit(mainpanel_instance, node,"Number of Buffers: ", 'numbuffers', 3, 0)

            addLineEdit(mainpanel_instance, node,"Heap Name: ", 'heapname', 3, 2)

            addLineEdit(mainpanel_instance, node,"Blocking Mode: ",'blocking_mode', 4, 0)

            addLineEdit(mainpanel_instance, node,"Reset msec Timeout: ", 'resetmsec_timeout', 4, 2)

        draw(mainpanel_instance, node, type_input)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our object to the factory '''
    factory.registerBlock("RealTimeThreadAsyncBridge", AsyncBridge, plugin_datastore)
    factory.registerBlock("AsyncBridge", AsyncBridge, plugin_datastore)



''' This module provides the base representation class for a MARTe2 python GAM '''
import copy

from martepy.marte2.config_object import MARTe2ConfigObject

class MARTe2GAM(MARTe2ConfigObject):
    """Object for configuring GAMs for MARTe2RealTimeThread instances"""
    plugin = "marte2"

    def __init__(self,
                    configuration_name: str = '+gam',
                    class_name: str = 'GAM',
                    input_signals: list = [],
                    output_signals: list = []
                ):
        super().__init__()
        self.configuration_name = configuration_name
        self.class_name = class_name
        self.input_signals = input_signals
        self.output_signals = output_signals

    @property
    def inputs(self):
        ''' Return the input signals as a fixed inputs attr '''
        return self.input_signals

    @property
    def outputs(self):
        ''' Return the output signals as a fixed outputs attr '''
        return self.output_signals


    def writeGamConfig(self, config_writer):
        ''' Write the GAM specific parameters '''
        raise NotImplementedError()

    def write(self, config_writer):
        ''' Write the total GAM configuration '''
        config_writer.startClass('+' + self.configuration_name.lstrip('+'), self.class_name)
        self.writeGamConfig(config_writer)
        if self.input_signals:
            config_writer.startSection('InputSignals')
            self.writeSignals(self.input_signals, config_writer)
            config_writer.endSection('InputSignals')
        if self.output_signals:
            config_writer.startSection('OutputSignals')
            self.writeSignals(self.output_signals, config_writer)
            config_writer.endSection('OutputSignals')
        config_writer.endSection('+' + self.configuration_name.lstrip('+'))

    def isOrderingRank0(self):
        ''' Always returns false, basically used to determine whether
        the GAM should always be placed first above all others in execution
        order (True) or not (False - solve as linked as to when order of
        execution should occur: i.e. after the input signals producer '''
        return False

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            myself = copy.deepcopy(self.serialize())
            theother = copy.deepcopy(other.serialize())
            myself['id'] = 0
            theother['id'] = 0
            if myself == theother:
                return True
        return False

    def serialize(self):
        ''' Serialize our GAM into a dictionary '''
        data = super().serialize()
        data['configuration_name'] = self.configuration_name
        data['inputsb'] = self.input_signals
        data['outputsb'] = self.output_signals
        data['class_name'] = self.class_name
        data['block_type'] = self.class_name
        data['comment'] = ''
        data['rank'] = False
        data['plugin'] = 'marte2'
        data['label'] = self.class_name
        data['parameters'] = {'Class name': self.class_name}
        data['content'] = {}
        data['inputs'] = []
        data['outputs'] = []
        data['id'] = id(self)
        data['pos_x'] = 0
        data['title'] = self.configuration_name + " (" + self.class_name + ")"
        # Need to have this match the format expected in GUI
        data['pos_y'] = 0
        return copy.deepcopy(data)

    def deserialize(self, data: dict, hashmap: dict={}, # pylint: disable=W1113, W0613, W0221
                    restore_id: bool=True) -> bool:
        ''' Deserialize our GAM object from a given dictionary '''
        super().deserialize(data, hashmap, restore_id)
        self.configuration_name = data['configuration_name']
        self.output_signals = data['outputsb']
        self.input_signals = data['inputsb']
        #self.class_name = data['class_name']
        # Now convert an input signal to what we store it as here
        # use the hidden _ versions of signals
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """

    def runOde(self):
        ''' For future use - pythonically simulate the GAM. '''
        # inputs is the input signals in their respective order outlined in the GAM.
        class_name = self.serialize()['class_name']
        # It is shown in this format so it is the format known by a user of this file.
        msg = f"This function has not been implemented for the GAM: {class_name}"
        raise NotImplementedError(msg)



"""This class acts as a storage database to store types and
packet definitions for later use in the application.
The packet definitions are also simply types however in this
context they could be SDN/UDP inputs/outputs to your system.

Types are stored in RDF format for the best computational storage possible.
"""

import re
import os
import shutil
from string import Template
import getpass
import copy
from datetime import datetime

from martepy.marte2.sorter import sortHeaders

STRUCT_PATTERN = re.compile(r'.*?typedef\s+struct\s*(\w*)\s*{([^}]*)}\s*(\w*)\s*;.*', re.DOTALL)
VERSION_PATTERN = r"_(\d+_\d+)\.h$"

class TypeException(Exception):
    """Exceptions to be raised specific to the type database"""

# Define field class
class Field: # pylint: disable=R0903
    ''' Type member definition '''
    def __init__(self, name, field_type, noelements=1, comment = ''):
        self.name = name
        self.type = field_type
        self.noelements = noelements
        self.comment = comment

    def __eq__(self, other):
        if not isinstance(other, Field):
            return False
        return (self.comment == other.comment and
                self.type == other.type and
                self.name == other.name and
                self.noelements == other.noelements)

# Define type class
class Type:
    ''' Class representation of a type '''
    def __init__(self, name, fundamental=False, file='', version='1_0', comments=''):
        self.name = name
        self.fundamental = fundamental
        self.fields = []
        self.file = file
        self.version = version
        self.comments = comments

    def addField(self, field):
        ''' Add a field '''
        self.fields.append(field)

    def __eq__(self, other):
        if not isinstance(other, Type):
            return False
        return (self.file == other.file and
                self.version == other.version and
                self.fundamental == other.fundamental and
                self.name == other.name and
                self.fields == other.fields)


class Constant(): # pylint: disable=R0903
    ''' Definition for any constant values - usually found in interpretting a type where
    the type has an array element and the array size is defined by a constant. '''
    def __init__(self, name, value, comments=''):
        self.name = name
        self.value = value
        self.comments = comments

class TypeDBv2():
    ''' The Type Database class '''
    def __init__(self):
        # In it's memory state it's best to store the data
        # in dictionary format as it gives us an O(1) efficieny
        # i.e. as the datastore increases in size, retrieving types does not
        # However since we store the types as objects in files then
        # That itself should have a O(N) efficiency.
        # This stores types
        self.types = {}
        # This stores definitions, mainly just useful as it allows multiple files
        # to use a shared set of definitions, as long as we parse those defining files first
        self.definitions = {}
        self.template_path = os.path.join(os.path.dirname(__file__), 'templates')

    def __eq__(self, other):
        if not isinstance(other, TypeDBv2):
            return False
        return self.types == other.types


    def _evaluateExpression(self, expression, constants):
        ''' Evaluate an expression in C++ '''
        def applyOperator(operators, values):
            ''' Apply C++ operations on a number '''
            while operators and len(values) > 1:
                operator = operators.pop()
                right = values.pop()
                left = values.pop()

                if operator == '<<':
                    values.append(left << right)
                elif operator == '>>':
                    values.append(left >> right)

        def isHexadecimal(s):
            ''' Evaluate whether it is a hexadecimal number '''
            try:
                # Attempt to convert the string to an integer with base 16
                int(s, 16)
                return True
            except ValueError:
                return False

        def parseTokens(tokens):
            ''' Token parser for interpretting a expression '''
            operators = []
            values = []

            for token in tokens:
                if token in constants:
                    values.append(constants[token].value)
                elif token == '(':
                    operators.append(token)
                elif token == ')':
                    while operators and operators[-1] != '(':
                        applyOperator(operators, values)
                    operators.pop()  # Remove the '('
                elif token in ['<<', '>>']:
                    while operators and operators[-1] in ['<<', '>>']:
                        applyOperator(operators, values)
                    operators.append(token)
                elif token.isdigit():
                    values.append(int(token))
                elif isHexadecimal(token):
                    values.append(int(token, 16))
                else:
                    raise ValueError(f"Invalid token: {token}")

            applyOperator(operators, values)

            return values[0] if not operators else values

        # Tokenize the expression
        tokens = re.findall(r'\b(?:<<|>>|\w+|\(|\))\b', expression)

        if not tokens:
            raise ValueError()
        # Evaluate the expression recursively
        result = parseTokens(tokens)

        return result

    def _getVersion(self, filepath):
        ''' Based on a files name, derive the version '''
        match = re.search(VERSION_PATTERN, os.path.basename(filepath))
        if match:
            return match.group(1)
        return '1_0'

    def _cleanContent(self, header_content):
        ''' Removes commented lines from the overall content '''
        filtered_list = [item for item in header_content if
                         not (item.startswith('#') or item == '\n')]
        overall_content = ''.join(filtered_list)
        return overall_content

    def _evaluateHeaderDefs(self, header_content):
        ''' This function evaluates any constant or numeric definitions in header contents '''
        for line in header_content:
            pattern = re.compile(r'^#define\s+(\w+)\s+(\S+)(.*)$')
            match = pattern.match(line)

            if match is None:
                continue

            if match:
                value = self._evaluateExpression(match.group(2), self.definitions)
                self.definitions[match.group(1)] = Constant(match.group(1), # name
                                                            value, # value
                                                            match.group(3)) # additional string

    def _parseHeader(self, header_content, file=''):
        ''' Read and parse a header files contents to a type '''
        # Regular expressions to match typedef structs
        self._evaluateHeaderDefs(header_content)

        # So now we have definitions captured, we can compute structs now
        # Let's start by driving efficiency by removing lines that start with #
        # Then let's squash our lines back together with any new line definitions,
        # this will help us get rid of comment fields later
        # Minimizing the content will help
        # Now let's first find all our typedefs

        matches = STRUCT_PATTERN.findall(self._cleanContent(header_content))

        def cleanString(input_string):
            ''' Clean our string line of any new lines and spaces unnecessarily '''
            # Remove new lines
            without_newlines = input_string.replace('\n', ' ')

            # Reduce multiple spaces to a single space
            without_multiple_spaces = re.sub(r'\s+', ' ', without_newlines)

            return without_multiple_spaces

        def removeCcomments(input_string):
            ''' Remove any C/C++ Comments '''
            pattern = r'\/\*[\s\S]*?\*\/'
            cleaned_string = re.sub(pattern, '', input_string)
            return cleaned_string

        def captureCcomments(input_string):
            ''' Capture the C/C++ Comments '''
            pattern = r'\/\*([\s\S]*?)\*\/'
            matches = re.findall(pattern, input_string)
            if matches:
                merged_string = '.'.join(matches)
                return merged_string
            return ''

        version = self._getVersion(file)
        # Finally create all our types in the header and update our dictionaries
        for match in matches:
            # This gets all our typedefs, for this level it just extracts the string for us
            # and the name if declared before or after the struct definition
            name = match[0] if match[2] == '' else match[2]
            new_type = Type(name, file=file, version = version)
            # Remove duplicate spaces and new line content and get us a flat string for our
            # definition only seperated by what we know
            # Now get our types
            # clean typedef content and split by ;
            for type_partial in cleanString(match[1]).split(';'):
                matches = re.findall(r'\s(\w+)\s(\w+)(?:\[(\w+)\])?\s*',
                                     removeCcomments(type_partial))
                if not matches:
                    continue
                noelements = 1 if matches[0][2] == '' else matches[0][2]
                try:
                    noelements = int(noelements)
                except ValueError:
                    # This means we have a string definition
                    try:
                        noelements = self._evaluateExpression(noelements,
                                                               self.definitions)
                    except ValueError as exc:
                        raise ValueError(f'''could not evaluate string of number of\n elements
 for field {matches[0][0]}, string is: {noelements}''') from exc
                type_def = matches[0][0].lower()
                type_def = 'uint32' if type_def == 'bitfield' else type_def
                type_def = 'float32' if type_def == 'float' else type_def
                new_type.fields.append(Field(matches[0][1],type_def,
                                             noelements,captureCcomments(type_partial)))
            self.types[name] = new_type

    def isFundamental(self, type_name):
        ''' Check if a given type is a fundamental type of MARTe2 '''
        return type_name.lower() in ["uint32", "uint64", "float32", "float64",
                                     "uint8", "int32", "int64", "int8", "int16"]

    def loadDb(self, directory_path):
        ''' Load a directory of types into the database '''
        header_paths = sortHeaders(directory_path)

        for header_path in header_paths:
            self.loadFile(header_path)

    def loadFile(self, file_path):
        ''' Load a type file '''
        with open(file_path, 'r', encoding='utf-8') as header_file:
            header_content = header_file.readlines()
            self._parseHeader(header_content, file=file_path)

    def getTypeByName(self, type_name):
        ''' Get a type based on it's name '''
        try:
            return self.types[type_name]
        except Exception as exc:
            raise TypeException(type_name) from exc

    def saveType(self, typedef, file_path):
        ''' Save the type to file '''
        self._updateStructInFile(file_path, typedef)

    # Function to generate the struct definition as a string
    def _generateStructDefinition(self, type_info):
        ''' Generate the struct definition for a type '''
        struct_def = "typedef struct {\n"
        for field in type_info.fields:
            if int(field.noelements) > 1:
                struct_def += f"    {field.type} {field.name}[{field.noelements}];\n"
            else:
                struct_def += f"    {field.type} {field.name};\n"
        struct_def += f"}} {type_info.name};"
        return struct_def

    # Function to update the struct in the file
    def _updateStructInFile(self, file_path, type_info):
        ''' Do an inplace saving of our type definition '''
        with open(file_path, 'r', encoding='utf-8') as file:
            content = file.read()

        # Regex to find the struct definition
        re_str = rf"typedef\s+struct\s*{{[^}}]*}}\s*{re.escape(type_info.name)}\s*;"
        struct_pattern = re.compile(re_str, re.DOTALL)

        # Generate the new struct definition
        new_struct_def = self._generateStructDefinition(type_info)

        # Replace the old struct with the new one
        updated_content = struct_pattern.sub(new_struct_def, content)

        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(updated_content)

    def toLibrary(self, output_path):
        ''' Output our type database as it's C++ compilable MARTe2 format '''
        paths= ''
        if not os.path.exists(output_path):
            os.mkdir(output_path)
        for packet_name, type_def in self.types.items():
            path = self._createSubfolder(output_path, packet_name)
            self._write(self._create(self.template_path, packet_name,
                                     type_def, 'cpp', self._cppSubstituteDict),
                                     path, f'{packet_name}.cpp')
            self._write(self._create(self.template_path, packet_name,
                                     type_def, 'h', self._hSubstituteDict),
                                     path, f'{packet_name}.h')
            self._write(self._create(self.template_path, packet_name,
                                     type_def, 'makefile', self._makefileSubstituteDict),
                                     path, 'Makefile.inc')
            self._write('include Makefile.inc', path, 'Makefile.gcc')
            if paths:
                paths += '        '  # adding indentation
            paths += f'{packet_name}.x\\\n'
        self._write(self._create(self.template_path, paths, '', 'root_makefile',
                                 self._rootSubstituteDict), output_path, 'Makefile.inc')
        self._write('export TARGET=x86-linux\n\ninclude Makefile.inc',
                    output_path, 'Makefile.x86-linux')
        return paths

    def _createSubfolder(self, output_path, packet_name) -> str:
        """Create the subfolder for the packet's type definition."""
        if not os.path.exists(path := os.path.join(output_path, packet_name)):
            os.mkdir(path)
        return path

    def _create(self, template_path, packet_name, type_def, extension, substitute_dict) -> str:
        """Create the file contents."""
        with open(os.path.join(template_path, f'{extension}_template.txt'), 'r',
                  encoding='utf-8') as fhand:
            src = Template(fhand.read())
            result = src.substitute(substitute_dict(packet_name, type_def))
            result = result.replace('£', '$').replace('Â','')
        fhand.close()
        return result

    def _write(self, definition, path, filename) -> None:
        """Write the file."""
        with open(os.path.join(path, filename), 'w', encoding='utf-8') as fhand:
            fhand.write(definition)
        fhand.close()

    def _cppSubstituteDict(self, packet_name, type_def) -> dict:
        """Create the substitution dictionary for the packet."""
        return {
            'type_name': f'{packet_name}Cls',
            'type_def': self._extractCppDef(packet_name, type_def),
            'header_name': f'{packet_name}.h',
            'file_name': f'{packet_name}.cpp',
        }

    def _hSubstituteDict(self, packet_name, type_def) -> dict:
        """Create the substitution dictionary for the packet."""
        return {
            'type_name_upper': f'{packet_name.upper()}_H_',
            'struct_def': self._extractHDef(packet_name, type_def),
            'file_name': f'{packet_name}.h',
        }

    def _makefileSubstituteDict(self, packet_name, _) -> dict:
        """Return the substitution dictionary for the library makefile."""
        return {
            'type_name_x': f'{packet_name}.x',
            'type_name': packet_name,
        }

    def _rootSubstituteDict(self, packet_name, _) -> dict:
        """Return the substitution dictionary for the root level makefile."""
        return {'type_list': packet_name}

    def _extractCppDef(self, packet_name, type_def) -> str:
        """Extract the cpp type definition."""
        declaration = ''
        entries = ''
        for field in type_def.fields:
            declaration += f'''DECLARE_CLASS_MEMBER({packet_name}, {field.name},
 {field.type}, "[{field.noelements}]", "");\n'''
            entries += f'&{packet_name}_{field.name}_introspectionEntry, '
        declaration += f'''static const MARTe::IntrospectionEntry*
 {packet_name}StructEntries[] = {{  {entries}0 }};\n'''
        declaration += f'DECLARE_STRUCT_INTROSPECTION({packet_name}, {packet_name}StructEntries)\n'
        return declaration

    def _extractHDef(self, packet_name, type_def) -> str:
        """Extract the header type definition."""
        declaration = f'    struct {packet_name} {{\n'
        for field in type_def.fields:
            declaration += f'        {field.type} {field.name};\n'
        declaration += '    };\n'
        return declaration

    def updateDb(self, type_db_filepath):
        ''' For each type, save them '''
        # Go through each type and save it
        for type_name, type_def in self.types.items():
            self.saveTypeDef(type_db_filepath, type_name, type_def)

    def incrementVersion(self, version):
        ''' Increment the version number of our type'''
        # Split the version string into major and minor parts
        major, minor = version.split('_')

        # Convert the minor part to an integer and increment it
        minor = int(minor) + 1

        # Reconstruct the version string
        new_version = f"{major}_{minor}"

        return new_version

    def updateFileVersion(self, file_path, new_version):
        ''' Update the version def in the file contents '''
        # Regular expression pattern to match the version in the filename
        pattern = r"_(\d+_\d+)\.h$"

        # Search for the pattern in the filename
        match = re.search(pattern, file_path)

        if match:
            current_version = match.group(1)

            # Create the new filename by replacing the old version with the new version
            new_file_path = file_path.replace(f"_{current_version}.h", f"_{new_version}.h")

            # Create the archive folder path
            archive_folder = os.path.join(os.path.dirname(file_path), "archive")

            # Ensure the archive folder exists
            os.makedirs(archive_folder, exist_ok=True)

            # Create the archived file path
            archived_file_path = os.path.join(archive_folder, os.path.basename(file_path))

            # Move the old version file to the archive folder
            shutil.move(file_path, archived_file_path)

            # Rename the file with the new version
            with open(archived_file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            with open(new_file_path, 'w', encoding='utf-8') as f:
                f.write(content)

            return new_file_path, archived_file_path
        return file_path

    def saveTypeDef(self, type_db_filepath, type_name, type_value):
        ''' Save a given type definition to file '''
        if not os.path.exists(self.types[type_name].file):
            # Save to file
            substitution = {'packetname': type_name,
                'version': self.types[type_name].version,
                'username': getpass.getuser(),
                'datetime': str(datetime.now())}
            with open(os.path.join(self.template_path, 'header.txt'), 'r',
                      encoding='utf-8') as fhand:
                src = Template(fhand.read())
                result = src.substitute(substitution)
            fhand.close()
            with open(os.path.join(type_db_filepath,
                                   f'{type_name}_{self.types[type_name].version}.h'),
                                   'w',
                                   encoding='utf-8') as outfile:
                outfile.write(result)
            outfile.close()
        else:
            old_type = copy.deepcopy(type_value)
            self.loadFile(type_value.file)
            new_type = self.types[type_name]
            self.types[type_name] = old_type
            if not old_type == new_type:
                # Has changes - increment
                version = self.types[type_name].version
                self.types[type_name].version = self.incrementVersion(version)
                # update filename
                version = self.types[type_name].version
                self.types[type_name].file = self.updateFileVersion(self.types[type_name].file,
                                                                    version)[0]
        self.saveType(self.types[type_name], self.types[type_name].file)



"""A module containing Serializable "Interface". We pretend its an abstract class
"""
from collections import OrderedDict


class Serializable():
    ''' Base class for serializable objects '''
    def __init__(self):
        """Default constructor automatically creates data which are
        common to any serializable object.
        In our case we create ``self.id`` which we use in every object in NodeEditor.
        """
        self.id = id(self)

    def serialize(self) -> OrderedDict:
        """Serialization method to serialize this class data into
        ``OrderedDict`` which can be easily stored in memory or file.

        :return: data serialized in ``OrderedDict``
        :rtype: ``OrderedDict``
        """
        raise NotImplementedError()

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        """Deserialization method which take data in python ``dict``
        format with helping `hashmap` containing references to existing entities.

        :param data: Dictionary containing serialized data
        :type data: ``dict``
        :param hashmap: Helper dictionary containing references (by id == key) to existing objects
        :type hashmap: ``dict``
        :param restore_id: True if we are creating new Sockets. False is useful when
        loading existing Sockets of which we want to keep the existing object's `id`.
        :type restore_id: bool
        :return: ``True`` if deserialization was successful, otherwise ``False``
        :rtype: ``bool``
        """
        raise NotImplementedError()



''' Pythonic representation of the directory resource '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2HttpDirectoryResource(MARTe2ConfigObject):
    ''' Pythonic representation of the directory resource '''
    def __init__(self,
                    configuration_name: str = 'WebRoot',
                    basedir = "",
                ):
        self.class_name = 'HttpDirectoryResource'
        self.basedir = basedir
        super().__init__()
        self.configuration_name = configuration_name

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        config_writer.startClass('+' + getname(self), self.class_name)
        config_writer.writeNode('BaseDir', f'"{self.basedir}"')
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['basedir'] = self.basedir
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.basedir = data['basedir']
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("HttpDirectoryResource", MARTe2HttpDirectoryResource, plugin_datastore)



''' Pythonic representation of GAMScheduler '''

import logging
logger = logging.getLogger(__name__)

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2GAMScheduler(MARTe2ConfigObject):
    """Object for configuring the default MARTe2 GAM scheduler"""

    def __init__(self,
                    configuration_name: str = '+Scheduler',
                    class_name: str = 'GAMScheduler',
                    timing_datasource_name: str = None,
                    maxcycles = 0
                ):
        super().__init__(configuration_name = configuration_name)
        self.class_name = class_name
        self.timing_datasource_name = timing_datasource_name
        self.maxcycles = maxcycles

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        config_writer.startClass('+' + getname(self), self.class_name)
        config_writer.writeNode('TimingDataSource', self.timing_datasource_name)
        config_writer.writeNode('MaxCycles', self.maxcycles)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['timing_datasource_name'] = self.timing_datasource_name
        res['maxcycles'] = self.maxcycles
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.timing_datasource_name = data["timing_datasource_name"]
        self.maxcycles = data["maxcycles"]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("GAMScheduler", MARTe2GAMScheduler, plugin_datastore)



''' Pythonic representative class of the EPICS Publisher Datasource '''
from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addComboEdit, addLineEdit, addInputSignalsSection

class EPICSPublisher(MARTe2DataSource):
    ''' Pythonic representation of the EPICS Publisher -
    using the one we have found to reliably work '''

    def __init__(self,
                    configuration_name: str = 'EPICSPublisher',
                    input_signals = [],
                    output_signals = [],
                    StackSize = 1048576,
                    CPUs = 0xff,
                    NumberOfBuffers=10,
                    IgnoreBufferOverrun=1,
                    DBR64CastDouble="yes",
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'EPICSCA::EPICSCAOutput',
                input_signals = input_signals,
            )
        self.stacksize = StackSize
        self.cpus = CPUs
        self.numberofbuffers = NumberOfBuffers
        self.ignorebufferoverrun = IgnoreBufferOverrun
        self.dbr64castdouble = DBR64CastDouble

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('StackSize', f'{self.stacksize}')
        if isinstance(self.cpus, str):
            if 'x' in self.cpus:
                config_writer.writeNode('CPUs', self.cpus)
            else:
                config_writer.writeNode('CPUs', hex(int(self.cpus)))
        else:
            config_writer.writeNode('CPUs', hex(self.cpus))
        config_writer.writeNode('NumberOfBuffers', f'{self.numberofbuffers}')
        config_writer.writeNode('IgnoreBufferOverrun', f'{self.ignorebufferoverrun}')
        config_writer.writeNode('DBR64CastDouble', f'{self.dbr64castdouble}')
        self.writeInputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialise the EPICS Publisher configuration '''
        res: dict = super().serialize()
        res['parameters']['Class name'] = 'EPICSPublisher'
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['cpus'] = self.cpus
        res['parameters']['numberofbuffers'] = self.numberofbuffers
        res['parameters']['ignorebufferoverrun'] = self.ignorebufferoverrun
        res['parameters']['dbr64castdouble'] = self.dbr64castdouble
        # We override this as the class name reference is ugly so we just
        # want to be called EPICSPublisher in the GUI
        res['label'] = "EPICSPublisher"
        res['block_type'] = 'EPICSPublisher'
        res['class_name'] = 'EPICSPublisher'
        res['title'] = f"{self.configuration_name} (EPICSPublisher)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the EPICS Publisher configuration '''
        res = super().deserialize(data, hashmap, restore_id)
        self.stacksize = data['parameters']["stacksize"]
        self.cpus = data['parameters']["cpus"]
        self.numberofbuffers = data['parameters']['numberofbuffers']
        self.ignorebufferoverrun = data['parameters']['ignorebufferoverrun']
        self.dbr64castdouble = data['parameters']['dbr64castdouble']
        return res

    def writeSignals(self, defs, config_writer):
        """Use MARTe1ConfigObject.writeSignals but remove any DataSources."""
        omit_keys = ('DataSource','Alias')
        signals = [(
                n,
                dict(d, **{'MARTeConfig': {
                        k: v for k, v in d['MARTeConfig'].items() if k not in omit_keys
                    }})
            ) for n, d in defs]

        super().writeSignals(signals, config_writer)

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addInputSignalsSection(mainpanel_instance, node, pack=False,
                               datasource=node.configuration_name, epics=True)

        addLineEdit(mainpanel_instance, node, "Stack Size: ", 'stacksize', 2, 0)

        if isinstance(node.parameters['cpus'], str):
            if 'x' in node.parameters['cpus']:
                node.parameters['cpus'] = node.parameters['cpus']
            else:
                node.parameters['cpus'] = hex(int(node.parameters['cpus']))
        else:
            node.parameters['cpus'] = hex(node.parameters['cpus'])

        addLineEdit(mainpanel_instance, node, "CPU Mask: ", 'cpus', 2, 2)

        addLineEdit(mainpanel_instance, node, "Number of Buffers: ", 'numberofbuffers', 3, 0)

        addComboEdit(mainpanel_instance, node, "Ignore Buffer Overruns?: ",
                     'ignorebufferoverrun', 3, 2, ['1','0'])

        addComboEdit(mainpanel_instance, node, "DBR64CastDouble: ",
                     'dbr64castdouble', 4, 0, ['yes','no'])

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our publisher with the factory '''
    factory.registerBlock("EPICSPublisher", EPICSPublisher, plugin_datastore)
    factory.registerBlock("EPICSCA::EPICSCAOutput", EPICSPublisher, plugin_datastore)



''' Pythonic representative class of the GAM Datasource '''
from martepy.marte2.datasource import MARTe2DataSource


class GAMDataSource(MARTe2DataSource):
    ''' Pythonic representation of the GAMDataSource '''
    def __init__(self,
                    configuration_name: str = 'DDB',
                    input_signals = [],
                    output_signals = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'GAMDataSource',
            )

    def writeDatasourceConfig(self, _):
        ''' Write nothing for our GAM configuration '''

    def serialize(self):
        ''' Serialize nothing '''
        res = super().serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize nothing additional '''
        res = super().deserialize(data, hashmap, restore_id)
        return res

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our object with the factory '''
    factory.registerBlock("GAMDataSource", GAMDataSource, plugin_datastore)



''' Pythonic representative class of the File Writer Datasource ''' 
from martepy.marte2.datasource import MARTe2DataSource


from martepy.marte2.qt_functions import addComboEdit, addLineEdit, addInputSignalsSection

class FileWriter(MARTe2DataSource):
    ''' Pythonic representation of the File Writer object '''
    def __init__(self,
                    configuration_name: str = 'FileWriter',
                    input_signals = [],
                    output_signals = [],
                    number_of_buffers: int = 100000,
                    cpu_mask: int = 0xFFFFFFFF,
                    stack_size: int = 100000000,
                    file_name: str = '',
                    overwrite: str = 'yes',
                    file_format: str = 'csv',
                    csv_separator: str = ',',
                    store_on_trigger: bool = False,
                    refreshcontent = 0,
                    numpretriggers = 0,
                    numposttriggers = 0
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'FileDataSource::FileWriter',
                input_signals = input_signals,
            )
        self.numberofbuffers = number_of_buffers
        self.cpumask = cpu_mask
        self.stacksize = stack_size
        self.filename = file_name
        self.overwrite = overwrite
        self.fileformat = file_format
        self.csvseparator = csv_separator
        self.storeontrigger = store_on_trigger
        self.refreshcontent = refreshcontent
        self.numberofpretriggers = numpretriggers
        self.numberofposttriggers = numposttriggers

    def writeDatasourceConfig(self, config_writer):
        ''' Write our datasource configuration '''
        def toIntString(string_val):
            return f'{int(string_val)}'
        config_writer.writeNode('NumberOfBuffers', toIntString(self.numberofbuffers))
        if isinstance(self.cpumask, str):
            if 'x' in self.cpumask:
                config_writer.writeNode('CPUMask', self.cpumask)
            else:
                config_writer.writeNode('CPUMask', hex(int(self.cpumask)))
        else:
            config_writer.writeNode('CPUMask', hex(self.cpumask))
        config_writer.writeNode('StackSize', toIntString(self.stacksize))
        config_writer.writeNode('Filename', f'"{self.filename}"')
        if isinstance(self.overwrite, str):
            if self.overwrite in ('yes', '1'):
                overwrite = 'yes'
            else:
                overwrite = 'no'
        else:
            overwrite = 'yes' if self.overwrite else 'no'
        config_writer.writeNode('Overwrite', f'"{overwrite}"')
        config_writer.writeNode('FileFormat', f'"{self.fileformat}"')
        if self.fileformat == 'csv':
            config_writer.writeNode('CSVSeparator', f'"{self.csvseparator}"')
        config_writer.writeNode('StoreOnTrigger', toIntString(self.storeontrigger))
        config_writer.writeNode('RefreshContent', toIntString(self.refreshcontent))
        config_writer.writeNode('NumberOfPreTriggers', toIntString(self.numberofpretriggers))
        config_writer.writeNode('NumberOfPostTriggers', toIntString(self.numberofposttriggers))
        self.writeInputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize our object '''
        res: dict = super().serialize()
        res['parameters']['numberofbuffers'] = self.numberofbuffers
        res['parameters']['cpumask'] = self.cpumask
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['storeontrigger'] = self.storeontrigger
        res['parameters']['refreshcontent'] = self.refreshcontent
        res['parameters']['numberofpretriggers'] = self.numberofpretriggers
        res['parameters']['numberofposttriggers'] = self.numberofposttriggers
        res['parameters']['filename'] = self.filename
        res['parameters']['overwrite'] = self.overwrite
        res['parameters']['fileformat'] = self.fileformat
        res['parameters']['csvseparator'] = self.csvseparator
        res['label'] = "FileWriter"
        res['block_type'] = 'FileWriter'
        res['class_name'] = 'FileWriter'
        res['title'] = f"{self.configuration_name} (FileWriter)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object to our class instance '''
        res: bool = super().deserialize(data, hashmap, restore_id)
        self.numberofbuffers = data['parameters']["numberofbuffers"]
        self.cpumask = data['parameters']["cpumask"]
        self.stacksize = data['parameters']["stacksize"]
        self.filename = data['parameters']["filename"]
        self.overwrite = data['parameters']["overwrite"]
        self.fileformat = data['parameters']["fileformat"]
        self.csvseparator = data['parameters']["csvseparator"]
        self.storeontrigger = data['parameters']["storeontrigger"]
        self.refreshcontent = data['parameters']["refreshcontent"]
        self.numberofpretriggers = data['parameters']["numberofpretriggers"]
        self.numberofposttriggers = data['parameters']["numberofposttriggers"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addInputSignalsSection(mainpanel_instance, node, False, datasource=node.configuration_name)

        addLineEdit(mainpanel_instance, node, "File name: ", 'filename', 2, 0)

        addComboEdit(mainpanel_instance, node, "File format: ",
                     'fileformat', 2, 2, ['CSV', 'Binary'])

        addLineEdit(mainpanel_instance, node, "Separator: ", 'csvseparator', 3, 0)

        addComboEdit(mainpanel_instance, node, "Overwrite: ", 'overwrite', 3, 2, ['Yes', 'No'])

        addComboEdit(mainpanel_instance, node, "Refresh Content: ",
                     'refreshcontent', 4, 0, ['1', '0'])

        if isinstance(node.parameters['cpumask'], str):
            if 'x' in node.parameters['cpumask']:
                node.parameters['cpumask'] = node.parameters['cpumask']
            else:
                node.parameters['cpumask'] = hex(int(node.parameters['cpumask']))
        else:
            node.parameters['cpumask'] = hex(node.parameters['cpumask'])

        addLineEdit(mainpanel_instance, node, "CPU Mask: ", 'cpumask', 4, 2)

        addLineEdit(mainpanel_instance, node, "Stack Size: ", 'stacksize', 5, 0)

        addComboEdit(mainpanel_instance, node, "Store on Trigger: ",
                     'storeontrigger', 5, 2, ['1', '0'])

        addLineEdit(mainpanel_instance, node, "Number of Buffers: ", 'numberofbuffers', 6, 0)

        addLineEdit(mainpanel_instance, node, "Number of Pretriggers: ",
                    'numberofpretriggers', 6, 2)

        addLineEdit(mainpanel_instance, node, "Number of Post Triggers: ",
                    'numberofposttriggers', 7, 0)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("FileDataSource::FileWriter", FileWriter, plugin_datastore)
    factory.registerBlock("FileWriter", FileWriter, plugin_datastore)



''' Import both the publisher and subscriber in one go '''

from martepy.marte2.datasources.epics.subscriber import EPICSSubscriber
from martepy.marte2.datasources.epics.publisher import EPICSPublisher

__all__ = [
        'EPICSSubscriber',
        'EPICSPublisher',
]



''' Pythonic representation of the SDN Publisher '''
from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addLineEdit, addInputSignalsSection

class SDNPublisher(MARTe2DataSource):
    ''' Pythonic representation of the SDN Publisher '''
    def __init__(self,
                    configuration_name: str = 'SDNPublisher',
                    input_signals = [],
                    output_signals = [],
                    topic: str = 'name',
                    interface: str = 'name',
                    address: str = '',
                    network_byte_order: int = 1,
                    source_port = '',
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'SDNPublisher',
                input_signals = input_signals,
            )
        self.topic = topic
        self.interface = interface
        self.port = source_port
        self.address = address
        self.byte_order = network_byte_order

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('Topic', f'{self.topic}')
        config_writer.writeNode('Interface', f'{self.interface}')
        config_writer.writeNode('NetworkByteOrder', f'{self.byte_order}')
        if self.address != '':
            config_writer.writeNode('Address', f'{self.address}')
        if self.port != '':
            config_writer.writeNode('SourcePort', f'{self.port}')
        self.writeInputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the object '''
        res: dict = super().serialize()
        res['parameters']['topic'] = self.topic
        res['parameters']['interface'] = self.interface
        res['parameters']['address'] = self.address
        res['parameters']['port'] = self.port
        res['parameters']['byte_order'] = self.byte_order
        res['label'] = "SDNPublisher"
        res['block_type'] = 'SDNPublisher'
        res['class_name'] = 'SDNPublisher'
        res['title'] = f"{self.configuration_name} (SDNPublisher)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object '''
        res = super().deserialize(data, hashmap, restore_id)
        self.topic = data['parameters']["topic"]
        self.interface = data['parameters']["interface"]
        self.byte_order = data['parameters']["byte_order"]
        self.address = data['parameters']["address"]
        self.port = data['parameters']["port"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addInputSignalsSection(mainpanel_instance,node, False, datasource=node.configuration_name)

        addLineEdit(mainpanel_instance,node,"Topic name: ", 'topic', 2, 0)

        addLineEdit(mainpanel_instance,node,"Interface: ", 'interface', 2, 2)

        addLineEdit(mainpanel_instance,node,"Address: ", 'address', 3, 0)

        addLineEdit(mainpanel_instance,node,"Source Port: ", 'port', 3, 2)

        addLineEdit(mainpanel_instance,node,"Network Byte Order: ", 'byte_order', 4, 0)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("SDNPublisher", SDNPublisher, plugin_datastore)



''' Mainly imports just the MARTe2Application instance and a basic GAM.'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.datasources import *
from martepy.marte2.gams import *
from martepy.marte2.objects import *

__all__ = [
    "MARTe2GAM",
    ]



''' Pythonic representative class of the Timing Datasource '''
from martepy.marte2.datasource import MARTe2DataSource


class TimingDataSource(MARTe2DataSource):
    ''' Pythonic representation of the Timing DataSource '''
    def __init__(self,
                    configuration_name: str = 'Timings',
                    input_signals = [],
                    output_signals = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'TimingDataSource'
            )

    def writeDatasourceConfig(self, config_writer):
        ''' Write nothing '''

    def serialize(self):
        ''' Serialize nothing - in future will probably need to have some definition '''
        res = super().serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize generic timing object '''
        res = super().deserialize(data, hashmap, restore_id)
        return res

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our object with the factory '''
    factory.registerBlock("TimingDataSource", TimingDataSource, plugin_datastore)



''' Pythonic representation of the SDN Subscriber '''
from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addComboEdit, addLineEdit, addOutputSignalsSection

class SDNSubscriber(MARTe2DataSource):
    ''' Pythonic representation of the SDN Subscriber '''
    def __init__(self,
                    configuration_name: str = 'SDNSubscriber',
                    input_signals = [],
                    output_signals = [],
                    execution_mode: str = 'IndependentThread',
                    topic: str = 'name',
                    interface: str = 'name',
                    internal_timeout: int = 0,
                    timeout: int = -1,
                    cpus: int = 0xFFFFFFFF,
                    ignore_timeout_error: int = 0,
                    class_name: str = 'SDNSubscriber'
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = class_name,
                output_signals = output_signals
            )
        self.topic = topic
        self.interface = interface
        self.execution_mode = execution_mode
        self.internal_timeout = internal_timeout
        self.timeout = timeout
        self.ignore_timeout_error = ignore_timeout_error
        self.cpus = cpus

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('Topic', f'"{self.topic}"')
        config_writer.writeNode('Interface', f'"{self.interface}"')
        config_writer.writeNode('IgnoreTimeoutError', f'"{int(self.ignore_timeout_error)}"')
        config_writer.writeNode('ExecutionMode', f'"{self.execution_mode}"')
        if self.execution_mode == "RealTimeThread":
            config_writer.writeNode('InternalTimeout', f'{self.internal_timeout}')
        else:
            config_writer.writeNode('Timeout', f'{self.timeout}')
        if isinstance(self.cpus, str):
            if 'x' in self.cpus:
                config_writer.writeNode('CPUs', self.cpus)
            else:
                config_writer.writeNode('CPUs', hex(int(self.cpus)))
        else:
            config_writer.writeNode('CPUs', hex(self.cpus))
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the configuration '''
        res: dict = super().serialize()
        res['parameters']['topic'] = self.topic
        res['parameters']['interface'] = self.interface
        res['parameters']['execution_mode'] = self.execution_mode
        res['parameters']['internal_timeout'] = self.internal_timeout
        res['parameters']['timeout'] = self.timeout
        res['parameters']['ignore_timeout_error'] = self.ignore_timeout_error
        res['parameters']['cpus'] = self.cpus
        res['label'] = "SDNSubscriber"
        res['block_type'] = 'SDNSubscriber'
        res['class_name'] = 'SDNSubscriber'
        res['title'] = f"{self.configuration_name} (SDNSubscriber)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object to our class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.topic = data['parameters']["topic"]
        self.interface = data['parameters']["interface"]
        self.execution_mode = data['parameters']["execution_mode"]
        self.internal_timeout = data['parameters']["internal_timeout"]
        self.timeout = data['parameters']["timeout"]
        self.ignore_timeout_error = data['parameters']["ignore_timeout_error"]
        self.cpus = data['parameters']["cpus"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addOutputSignalsSection(mainpanel_instance, node, False)

        addLineEdit(mainpanel_instance, node,"Topic: ",'topic', 2, 0)

        addLineEdit(mainpanel_instance, node,"Interface: ",'interface', 2, 2)

        addComboEdit(mainpanel_instance, node,"Execution mode: ", 'execution_mode',
                     3, 0, ['IndependentThread', 'RealTimeThread'])

        addLineEdit(mainpanel_instance, node,"Timeout: ",'timeout', 3, 2)

        addComboEdit(mainpanel_instance, node,"Ignore timeout error: ",
                     'ignore_timeout_error', 4, 0, ['0', '1'])

        addLineEdit(mainpanel_instance, node,"Internal timeout: ",'internal_timeout', 4, 2)
        if isinstance(node.parameters['cpus'], str):
            if 'x' in node.parameters['cpus']:
                node.parameters['cpus'] = node.parameters['cpus']
            else:
                node.parameters['cpus'] = hex(int(node.parameters['cpus']))
        else:
            node.parameters['cpus'] = hex(node.parameters['cpus'])
        addLineEdit(mainpanel_instance, node,"CPU mask: ",'cpus', 5, 0)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("SDNSubscriber", SDNSubscriber, plugin_datastore)



''' This example shows how to create a basic setup of a HTTP Service in MARTe2 '''

from martepy.marte2.objects.http import *
from martepy.marte2.objects.message import MARTe2Message

def httpService(directory,app=None):
    """Fixed definition of HTTP Service however because the datagramtest
    repo has a dynamic definition this is possible to make dynamic here also"""
    # First thing to do is define the messages we might send to the
    # StateMachine (When working of course!)
    Messages = [MARTe2Message("+GOTOSTATE1","StateMachine","GOTOSTATE1"),
                MARTe2Message("+GOTOSTATE2","StateMachine","GOTOSTATE2")]

    # Object Browser Definition
    ObjectBrowser = MARTe2HTTPObjectBrowser('+ObjectBrowse','/')

    # ResourcesHtml Definition
    ResourcesHtml = MARTe2HttpDirectoryResource('+ResourcesHtml',
                                                f'{directory}/Resources/HTTP/')

    MessageInterface = MARTe2HttpMessageInterface('+HttpMessageInterface',Messages)

    Objectlist = [ObjectBrowser,ResourcesHtml,MessageInterface]

    HTTPBrowser = MARTe2HTTPObjectBrowser('+WebRoot','.',Objectlist)

    # Add WebServer
    service = MARTe2HttpService('+WebService')
    # Add this to application
    app.add(externals=[HTTPBrowser] + [service])



''' Pythonic representative class of the FileReader and Writer created that flushes with each
write to ensure the data is written without breakages in columns during CSV writing Datasource '''

from functools import partial

from PyQt5.QtWidgets import (
    QPushButton,
    QWidget,
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QFileDialog
)

from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import (addComboEdit,
                                         addInputSignalsSection,
                                         addOutputSignalsSection,
                                         addLineEdit,
                                         paraChange)

class RFileWriter(MARTe2DataSource):
    ''' Pythonic representation of Flushing FileWriter '''
    def __init__(self,
                    configuration_name: str = 'FileWriter',
                    input_signals = [],
                    output_signals = [],
                    numberofbuffers: int = 100000,
                    cpumask: int = 0xFFFFFFFF,
                    stacksize: int = 100000000,
                    filename: str = None,
                    overwrite: bool = True,
                    fileformat: str = 'csv',
                    csvseparator: str = ',',
                    storeontrigger: bool = False,
                    refreshcontent = 0,
                    numberofpretriggers = 0,
                    numberofposttriggers = 0
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'RFileDataSource::FileWriter',
                input_signals = input_signals
            )
        self.numberofbuffers = numberofbuffers
        self.cpumask = cpumask
        self.stacksize = stacksize
        self.filename = filename
        self.overwrite = overwrite
        self.fileformat = fileformat
        self.csvseparator = csvseparator
        self.storeontrigger = storeontrigger
        self.refreshcontent = refreshcontent
        self.numberofpretriggers = numberofpretriggers
        self.numberofposttriggers = numberofposttriggers

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource config '''
        def toIntString(string_val):
            return f'{int(string_val)}'

        config_writer.writeNode('NumberOfBuffers', toIntString(self.numberofbuffers))
        if isinstance(self.cpumask, str):
            if 'x' in self.cpumask:
                config_writer.writeNode('CPUMask', self.cpumask)
            else:
                config_writer.writeNode('CPUMask', hex(int(self.cpumask)))
        else:
            config_writer.writeNode('CPUMask', hex(self.cpumask))
        config_writer.writeNode('StackSize', toIntString(self.stacksize))
        config_writer.writeNode('Filename', f'"{self.filename}"')
        if isinstance(self.overwrite, str):
            if self.overwrite in ('yes', '1'):
                overwrite = 'yes'
            else:
                overwrite = 'no'
        else:
            overwrite = 'yes' if self.overwrite else 'no'
        config_writer.writeNode('Overwrite', f'"{overwrite}"')
        config_writer.writeNode('FileFormat', f'"{self.fileformat}"')
        if self.fileformat == 'csv':
            config_writer.writeNode('CSVSeparator', f'"{self.csvseparator}"')
        config_writer.writeNode('StoreOnTrigger', toIntString(self.storeontrigger))
        config_writer.writeNode('RefreshContent', toIntString(self.refreshcontent))
        config_writer.writeNode('NumberOfPreTriggers', toIntString(self.numberofpretriggers))
        config_writer.writeNode('NumberOfPostTriggers', toIntString(self.numberofposttriggers))
        self.writeInputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the datasource configuration '''
        res: dict = super().serialize()
        res['parameters']['numberofbuffers'] = self.numberofbuffers
        res['parameters']['cpumask'] = self.cpumask
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['storeontrigger'] = self.storeontrigger
        res['parameters']['refreshcontent'] = self.refreshcontent
        res['parameters']['numberofpretriggers'] = self.numberofpretriggers
        res['parameters']['numberofposttriggers'] = self.numberofposttriggers
        res['parameters']['filename'] = self.filename
        res['parameters']['overwrite'] = self.overwrite
        res['parameters']['fileformat'] = self.fileformat
        res['parameters']['csvseparator'] = self.csvseparator
        res['label'] = "FileWriter"
        res['block_type'] = 'FileWriter'
        res['class_name'] = 'FileWriter'
        res['title'] = f"{self.configuration_name} (FileWriter)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the datasource configuration '''
        res: bool = super().deserialize(data, hashmap, restore_id)
        deserialize_props = ["numberofbuffers",
                             "cpumask",
                             "stacksize",
                             "filename",
                             "overwrite",
                             "fileformat",
                             "csvseparator",
                             "storeontrigger",
                             "refreshcontent",
                             "numberofpretriggers",
                             "numberofposttriggers"]

        self.deserializeProperties(data, deserialize_props)
        return res

    def deserializeProperties(self, data, properties):
        '''Simplifies the deserialization of a list of properties where this becomes
        long.'''
        for property_name in properties:
            if property_name in list(data['parameters'].keys()):
                setattr(self, property_name, data['parameters'][property_name])

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addInputSignalsSection(mainpanel_instance, node, False, datasource=node.configuration_name)

        addLineEdit(mainpanel_instance, node, "File name: ", 'filename', 2, 0)

        addComboEdit(mainpanel_instance, node, "File format: ", 'fileformat',
                     2, 2, ['CSV', 'Binary'])

        addLineEdit(mainpanel_instance, node, "Separator: ", 'csvseparator', 3, 0)

        addComboEdit(mainpanel_instance, node, "Overwrite: ", 'overwrite', 3, 2, ['Yes', 'No'])

        addComboEdit(mainpanel_instance, node, "Refresh Content: ", 'refreshcontent',
                     4, 0, ['1', '0'])

        if isinstance(node.parameters['cpumask'], str):
            if 'x' in node.parameters['cpumask']:
                node.parameters['cpumask'] = node.parameters['cpumask']
            else:
                node.parameters['cpumask'] = hex(int(node.parameters['cpumask']))
        else:
            node.parameters['cpumask'] = hex(node.parameters['cpumask'])

        addLineEdit(mainpanel_instance, node, "CPU Mask: ", 'cpumask', 4, 2)

        addLineEdit(mainpanel_instance, node, "Stack Size: ", 'stacksize', 5, 0)

        addComboEdit(mainpanel_instance, node, "Store on Trigger: ", 'storeontrigger',
                     5, 2, ['1', '0'])

        addLineEdit(mainpanel_instance, node, "Number of Buffers: ", 'numberofbuffers', 6, 0)

        addLineEdit(mainpanel_instance, node, "Number of Pretriggers: ",
                    'numberofpretriggers', 6, 2)

        addLineEdit(mainpanel_instance, node, "Number of Post Triggers: ",
                    'numberofposttriggers', 7, 0)


class RFileReader(MARTe2DataSource):
    ''' Pythonic representation of Flushing FileReader '''
    def __init__(self,
                    configuration_name: str = 'FileReader',
                    filename = "",
                    fileformat = "csv",
                    csvseparator = ",",
                    interpolate = "no",
                    eof = "Rewind",
                    preload = "yes",
                    input_signals = [],
                    output_signals = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'RFileDataSource::FileReader'
            )
        self.output_signals = output_signals
        self.filename = filename
        self.fileformat = fileformat
        self.csvseparator = csvseparator
        self.interpolate = interpolate
        self.eof = eof
        self.preload = preload

    def writeDatasourceConfig(self, config_writer):
        ''' Write our datasource configuration '''
        config_writer.writeNode('Filename', f'"{self.filename}"')
        config_writer.writeNode('FileFormat', f'"{self.fileformat}"')
        config_writer.writeNode('CSVSeparator', f'"{self.csvseparator}"')
        config_writer.writeNode('Interpolate', f'"{self.interpolate}"')
        config_writer.writeNode('EOF', f'"{self.eof}"')
        config_writer.writeNode('Preload', f'"{self.preload}"')
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        res: dict = super().serialize()
        res['parameters']['filename'] = self.filename
        res['parameters']['fileformat'] = self.fileformat
        res['parameters']['interpolate'] = self.interpolate
        res['parameters']['csvseparator'] = self.csvseparator
        res['parameters']['eof'] = self.eof
        res['parameters']['preload'] = self.preload

        # Override name
        res['label'] = "FileReader"
        res['block_type'] = 'FileReader'
        res['class_name'] = 'FileReader'
        res['title'] = f"{self.configuration_name} (FileReader)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        res = super().deserialize(data, hashmap, restore_id)
        self.filename = data['parameters']["filename"]
        self.fileformat = data['parameters']["fileformat"]
        self.interpolate = data['parameters']["interpolate"]
        self.csvseparator = data['parameters']["csvseparator"]
        self.eof = data['parameters']["eof"]
        self.preload = data['parameters']["preload"]
        return res

    def writeSignals(self, defs, config_writer):
        """Use MARTe1ConfigObject.writeSignals but remove any DataSources."""
        omit_keys = ('DataSource','Alias')
        signals = [(
                n,
                dict(d, **{'MARTeConfig': {
                        k: v for k, v in d['MARTeConfig'].items() if k not in omit_keys
                    }})
            ) for n, d in defs]

        super().writeSignals(signals, config_writer)

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addOutputSignalsSection(mainpanel_instance, node, False)

        # Special file input
        wgt_label = QLabel("File name: ")
        wgt_field = QWidget()
        hlay = QHBoxLayout()
        wgt_field.setLayout(hlay)
        wgt_line = QLineEdit(mainpanel_instance)
        wgt_button = QPushButton("Browse")
        hlay.addWidget(wgt_line)
        hlay.addWidget(wgt_button)
        def filedialog():
            # Open a file dialog to select a file
            file_dialog = QFileDialog()
            file_dialog.setFileMode(QFileDialog.ExistingFile)
            if file_dialog.exec_():
                file_paths = file_dialog.selectedFiles()
                if file_paths:
                    # Read the text content from the selected file
                    file_path = file_paths[0]
                    wgt_line.setText(file_path)

        wgt_button.clicked.connect(filedialog)
        wgt_line.setText(str(node.parameters['filename']))
        wgt_line.textChanged.connect(partial(paraChange, wgt_line, node, 'filename'))
        mainpanel_instance.configbarBox.addWidget(wgt_label, 2, 0)
        mainpanel_instance.configbarBox.addWidget(wgt_field, 2, 1)

        addComboEdit(mainpanel_instance, node,"File format: ", 'fileformat',
                     2, 2, ['csv', 'binary'])

        addComboEdit(mainpanel_instance, node,"Interpolate: ", 'interpolate',
                     3, 0, ['yes', 'no'])

        addLineEdit(mainpanel_instance, node,"Separator: ",'csvseparator', 3, 2)

        addComboEdit(mainpanel_instance, node,"Preload: ", 'preload', 4, 0, ['yes', 'no'])

        addComboEdit(mainpanel_instance, node,"EOF: ", 'eof', 4, 2, ['Rewind', 'Last', 'Error'])

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the objects with the factory '''
    factory.registerBlock("RFileDataSource::FileReader", RFileReader, plugin_datastore)
    factory.registerBlock("RFileReader", RFileReader, plugin_datastore)
    factory.registerBlock("RFileDataSource::FileWriter", RFileWriter, plugin_datastore)
    factory.registerBlock("RFileWriter", RFileWriter, plugin_datastore)



''' Easy method to import all the state machine items '''

from martepy.marte2.objects.statemachine.event import MARTe2StateMachineEvent
from martepy.marte2.objects.statemachine.machine import MARTe2StateMachine

__all__ = [
        'MARTe2StateMachineEvent',
        'MARTe2StateMachine'
    ]



''' Pythonic representation of the UDP Receiver '''
from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addComboEdit, addOutputSignalsSection, addLineEdit


class UDPReceiver(MARTe2DataSource):
    ''' Pythonic representation of the UDP Receiver '''
    def __init__(self,
                    configuration_name: str = 'UDPReceiver',
                    input_signals = [],
                    output_signals = [],
                    address: str = '127.0.0.1',
                    interfaceaddress: str = '0.0.0.0',
                    stacksize = 10000000,
                    executionmode: str = 'IndependentThread',
                    timeout: int = 0,
                    port: int = 40442,
                    cpu_mask: int = 0xFFFFFFFF,
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'UDP::UDPReceiver',
                output_signals = output_signals
            )
        self.executionmode = executionmode
        self.timeout = timeout
        self.port = port
        self.cpu_mask = cpu_mask
        self.address = address
        self.interfaceaddress = interfaceaddress
        self.stacksize = stacksize

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('ExecutionMode', f'"{self.executionmode}"')
        config_writer.writeNode('Timeout', f'{self.timeout}')
        config_writer.writeNode('Port', f'{self.port}')
        if isinstance(self.cpu_mask, str):
            if 'x' in self.cpu_mask:
                config_writer.writeNode('CPUMask', self.cpu_mask)
            else:
                config_writer.writeNode('CPUMask', hex(int(self.cpu_mask)))
        else:
            config_writer.writeNode('CPUMask', hex(self.cpu_mask))
        config_writer.writeNode('Address', f'"{self.address}"')
        config_writer.writeNode('InterfaceAddress', f'"{self.interfaceaddress}"')
        config_writer.writeNode('StackSize', f'{self.stacksize}')
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the object '''
        res: dict = super().serialize()
        res['parameters']['executionmode'] = self.executionmode
        res['parameters']['timeout'] = self.timeout
        res['parameters']['port'] = self.port
        res['parameters']['address'] = self.address
        res['parameters']['interfaceaddress'] = self.interfaceaddress
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['cpu_mask'] = self.cpu_mask
        res['label'] = "UDPReceiver"
        res['block_type'] = 'UDPReceiver'
        res['class_name'] = 'UDPReceiver'
        res['title'] = f"{self.configuration_name} (UDPReceiver)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object to the class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.executionmode = data['parameters']["executionmode"]
        self.timeout = data['parameters']["timeout"]
        self.port = data['parameters']["port"]
        self.cpu_mask = data['parameters']["cpu_mask"]
        self.address = data['parameters']["address"]
        self.interfaceaddress = data['parameters']["interfaceaddress"]
        self.stacksize = data['parameters']["stacksize"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addOutputSignalsSection(mainpanel_instance, node)

        addLineEdit(mainpanel_instance, node, "Port: ", 'port', 2, 0)

        addLineEdit(mainpanel_instance, node, "Timeout: ", 'timeout', 2, 2)

        addLineEdit(mainpanel_instance, node, "Address: ", 'address', 3, 0)

        addLineEdit(mainpanel_instance, node, "Interface Address: ", 'interfaceaddress', 3, 2)

        addComboEdit(mainpanel_instance, node, "Execution Mode: ", 'executionmode',
                     4, 0, ['IndependentThread', 'RealTimeThread'])

        addLineEdit(mainpanel_instance, node, "Stack Size: ", 'stacksize', 4, 2)
        if isinstance(node.parameters['cpu_mask'], str):
            if 'x' in node.parameters['cpu_mask']:
                node.parameters['cpu_mask'] = node.parameters['cpu_mask']
            else:
                node.parameters['cpu_mask'] = hex(int(node.parameters['cpu_mask']))
        else:
            node.parameters['cpu_mask'] = hex(node.parameters['cpu_mask'])
        addLineEdit(mainpanel_instance, node, "CPU Mask: ", 'cpu_mask', 5, 0)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("UDPReceiver", UDPReceiver, plugin_datastore)



''' Import UDP Receiver and Sender all in one go '''

from martepy.marte2.datasources.udp.sender import UDPSender
from martepy.marte2.datasources.udp.receiver import UDPReceiver

__all__ = [
        'UDPSender',
        'UDPReceiver',
]



# pylint: disable=C0103
"""A module for creating MARTe configuration files in the output"""

import warnings

class ConfigWriter:
    ''' The Configuration Writer base class that is used by all MARTe2 objects 
    to write their configuration representation. '''
    def __init__(self, tab_text='    '):
        self.sectionStack = []
        self.tab = 0
        self.tab_text = tab_text

    def _output(self, line):
        ''' Internally used to produce the line output '''
        raise NotImplementedError('Please use a child class of ConfigWriter')

    def setTab(self, newtab, override=False):
        ''' Set the tab indentation currently.'''
        if len(self.sectionStack) != 0:
            if override:
                msg = f"Setting tab, section stack isn't empty! Tab was {self.tab}, now {newtab}"
                warnings.warn(msg, UserWarning)
            else:
                msg = f'''Asked to setTab({newtab}) when stack isn't empty.
Tab = {self.tab}, Stack = {self.sectionStack[-1]}'''
                raise IndexError(msg)
        self.tab = newtab

    def startSection(self, name):
        ''' Start a new indentation section encapsulated with {} but not necessarily
        a new class - does not enforce the class definition line. '''
        self.writeBareLine(f'{name} = {{')
        self.tab += 1
        self.sectionStack.append(name)

    def writeNode(self, name, value):
        ''' Write parameter or basic line'''
        self.writeBareLine(f'{name} = {value}')

    def writeMARTe2Vector(self, name, data, formatAsFloat=True):
        ''' Write a MARTe2 vector/array. '''
        if formatAsFloat:
            datastring = ' '.join(['%g' % d for d in data]) # pylint: disable=C0209
        else:
            datastring = ' '.join([f'{d}' for d in data])
        self.writeNode(name, f'{{ {datastring} }}')

    def writeBareLine(self, line):
        ''' Write a basic line as given with the current set tab indentation '''
        prefix = self.getPrefix()
        self._output(f'{prefix}{line}')

    def startClass(self, name, typename):
        ''' Begin writing a class, encapsulating it with { and enforcing the Class
        definition is written. '''
        self.startSection(name)
        self.writeNode('Class', typename)

    def endSection(self, name):
        ''' End the current section - must be the current section and must match in starting
        name. '''
        if self.sectionStack[-1] == name:
            self.tab -= 1
            self.writeBareLine('}')
            self.sectionStack.pop()
        else:
            raise IndexError(f'Failed trying to close section "{name}"')

    def getPrefix(self):
        ''' Get the current tab index in number of spaces.'''
        if self.tab < 0:
            raise IndexError(f'Tab Error, tab is set to: {self.tab}')
        return ''.join([self.tab_text]*self.tab)

class StringConfigWriter(ConfigWriter):
    '''The derived class which actually outputs when used a string output and returns the
    MARTe2 object as a string object. '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.lines = []
    def _output(self, line):
        ''' Append to our existing lines. '''
        self.lines.append(line)
    def __repr__(self):
        ''' Give a full output string format of the configuration object '''
        return '\n'.join(self.toLines())
    def toString(self):
        ''' Produce the object as a string '''
        return str(self)
    def toLines(self):
        ''' Return the currently defined lines. '''
        return self.lines



''' Pythonic representation of the State Machine Event Object '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.marte2.objects.message import MARTe2Message
from martepy.functions.extra_functions import getname

class MARTe2StateMachineEvent(MARTe2ConfigObject):
    ''' Pythonic representation of the State Machine Event Object '''
    def __init__(self,
                    configuration_name: str = 'GOTOSTATE1',
                    nextstate = "",
                    nextstateerror = "",
                    timeout = 0,
                    messages = [],
                ):
        super().__init__(
            )
        self.class_name = "StateMachineEvent"
        self.configuration_name = configuration_name.upper()
        self.nextstate = nextstate.upper()
        self.nextstateerror = nextstateerror.upper()
        self.timeout = timeout
        self.messages = messages

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            attributes = ['configuration_name', 'nextstate',
                          'nextstateerror', 'timeout', 'messages']
            for item in attributes:
                if not getattr(self, item) == getattr(other, item):
                    return False
            return True
        return False

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        def sanitize(value):
            return value.replace('"','')

        config_writer.startClass('+' + getname(self),"StateMachineEvent")
        config_writer.writeNode('NextState', f'"{sanitize(self.nextstate)}"')
        if self.nextstateerror != "":
            config_writer.writeNode('NextStateError', f'"{sanitize(self.nextstateerror)}"')
        config_writer.writeNode('Timeout', f'{self.timeout}')
        for message in self.messages:
            message.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['nextstate'] = self.nextstate
        res['nextstateerror'] = self.nextstateerror
        res['timeout'] = self.timeout
        res["messages"] = [a.serialize() for a in self.messages]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data['class_name']
        self.nextstate = data['nextstate']
        self.nextstateerror = data['nextstateerror']
        self.timeout = data['timeout']
        self.messages = [MARTe2Message().deserialize(a,factory=factory) for a in data['messages']]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("StateMachineEvent", MARTe2StateMachineEvent, plugin_datastore)
